from abc import ABC, abstractmethod

import numpy as np

from boltzgen.data.data import Structure


class StaticFilter(ABC):
    """Base class for structure filters."""

    @abstractmethod
    def filter(self, structure: Structure) -> np.ndarray:
        """Filter chains in a structure.

        Parameters
        ----------
        structure : Structure
            The structure to filter chains from.

        Returns
        -------
        np.ndarray
            The chains to keep, as a boolean mask.

        """
        raise NotImplementedError



import io
import warnings
from pathlib import Path
import random
from typing import List
import subprocess
import re
import biotite
import hydride
from sklearn.cluster import DBSCAN
from Bio import PDB
from biotite import structure
from Bio.Seq import Seq


from matplotlib import pyplot as plt
import matplotlib as mpl

mpl.rcParams["figure.max_open_warning"] = 100

from Bio import Align

from boltzgen.data.rmsd_computation import get_true_coordinates
from boltzgen.model.loss.diffusion import weighted_rigid_align
from boltzgen.task.predict.data_from_generated import collate

import torch
import numpy as np
import pandas as pd
from tqdm import tqdm

from boltzgen.data import const
from boltzgen.data.data import Structure, biotite_array_from_feat
from boltzgen.data.write.mmcif import to_mmcif
from boltzgen.model.loss.validation import factored_lddt_loss, compute_subset_rmsd

from biotite.structure.sasa import sasa
from biotite.structure.info import vdw_radius_single, vdw_radius_protor
import biotite.structure.io.pdbx as pdbx
import biotite.structure.io.pdb as pdbio

TARGET_ID_RE = re.compile(
    r"^(?:(?:sample\d+_|batch\d+_|rank\d+_)+)?([^_]+)(?:_[^_]+)*?(?:_(?:gen))*$"
)


def _load_stack(path):
    path = Path(path)
    suffix = path.suffix.lower()
    if suffix in {".cif", ".mmcif"}:
        cif_file = pdbx.CIFFile.read(str(path))
        stack = pdbx.get_structure(cif_file, use_author_fields=False)
    elif suffix in {".pdb", ".ent"}:
        pdb_file = pdbio.PDBFile.read(str(path))
        stack = pdbio.get_structure(pdb_file, model=None)
    else:
        raise ValueError(f"Unsupported structure file extension: {suffix}")

    return stack


def compute_rmsd(atom_coords: torch.Tensor, pred_atom_coords: torch.Tensor):
    rmsd, _ = compute_subset_rmsd(
        atom_coords,
        pred_atom_coords,
        atom_mask=torch.ones_like(atom_coords[..., 0]),
        align_weights=torch.ones_like(atom_coords[..., 0]),
        subset_mask=torch.ones_like(atom_coords[..., 0]),
        multiplicity=1,
    )
    return rmsd


def make_histogram(
    df,
    column_name: str,
):
    data = df[column_name].dropna()
    fig, ax = plt.subplots(figsize=(6, 4))

    ax.hist(data, bins=50, alpha=0.7, edgecolor="black", linewidth=0.5)
    ax.axvline(data.mean(), color="red", linestyle="dashed", linewidth=1)

    ax.set_title(
        f"{column_name.replace('_', ' ').capitalize()} Distribution", fontsize=12
    )
    ax.set_xlabel(column_name.replace("_", " "), fontsize=10)
    ax.set_ylabel("Count", fontsize=10)

    ax.grid(True, axis="y", linestyle="--", linewidth=0.5, alpha=0.6)
    ax.tick_params(axis="both", which="major", labelsize=8)

    plt.tight_layout()
    return fig


def get_best_folding_sample(folded):
    confidence = 0.8 * folded["design_to_target_iptm"] + 0.2 * folded["design_ptm"]
    best_idx = np.argmax(confidence)
    
    # TODO: remove the "if k in folded"
    best_sample = {
        k: folded[k][best_idx] for k in const.eval_keys_confidence if k in folded
    }
    best_sample["coords"] = folded["coords"][best_idx]
    return best_sample


def get_fold_metrics(
    feat,
    folded,
    compute_lddts=True,
    prefix="",
):
    batch = collate([feat])
    diffusion_samples = batch["coords"].shape[0]
    best_sample = get_best_folding_sample(folded)

    # Compute RMSDs
    rmsd_out = get_true_coordinates(
        batch=batch,
        out={"sample_atom_coords": torch.from_numpy(best_sample["coords"])},
        diffusion_samples=1,
        symmetry_correction=False,
        protein_lig_rmsd=True,
    )
    true_coords_resolved_mask = rmsd_out["true_coords_resolved_mask"]

    # Add to metrics dictionary
    metrics = {}
    metrics["rmsd"] = rmsd_out.get("rmsd").item()
    metrics["rmsd_design"] = rmsd_out.get("rmsd_design").item()
    metrics["rmsd_target"] = rmsd_out.get("rmsd_target").item()
    metrics["rmsd_design_target"] = rmsd_out.get("rmsd_design_target").item()
    metrics["target_aligned_rmsd_design"] = rmsd_out.get(
        "target_aligned_rmsd_design"
    ).item()
    metrics["rmsd<2.5"] = bool(metrics["rmsd"] <= 2.5)
    metrics["target_aligned<2.5"] = bool(metrics["target_aligned_rmsd_design"] <= 2.5)
    metrics["designability_rmsd_2"] = bool(metrics["rmsd_design"] <= 2.0)
    metrics["designability_rmsd_4"] = bool(metrics["rmsd_design"] <= 4.0)

    # Comput LDDTs
    if compute_lddts:
        all_lddt_dict, _ = factored_lddt_loss(
            feats=batch,
            atom_mask=true_coords_resolved_mask,
            true_atom_coords=batch["coords"],
            pred_atom_coords=torch.from_numpy(best_sample["coords"]),
            multiplicity=diffusion_samples,
            exclude_ions=False,
        )
        metrics.update({f"lddt_{k}": v.max().item() for k, v in all_lddt_dict.items()})
        metrics["designability_lddt_60"] = bool(metrics["lddt_intra_design"] >= 0.6)
        metrics["designability_lddt_65"] = bool(metrics["lddt_intra_design"] >= 0.65)
        metrics["designability_lddt_70"] = bool(metrics["lddt_intra_design"] >= 0.7)
        metrics["designability_lddt_75"] = bool(metrics["lddt_intra_design"] >= 0.75)
        metrics["designability_lddt_80"] = bool(metrics["lddt_intra_design"] >= 0.8)
        metrics["designability_lddt_85"] = bool(metrics["lddt_intra_design"] >= 0.85)
        metrics["designability_lddt_90"] = bool(metrics["lddt_intra_design"] >= 0.9)

    # metrics without prefix (backbone only is the same as all atom)
    # TODO: remove the "if k in best_sample"
    confs = {k: best_sample[k] for k in const.eval_keys_confidence if k in best_sample}
    confs["min_interaction_pae<1.5"] = bool(confs["min_interaction_pae"] <= 1.5)
    confs["min_interaction_pae<2"] = bool(confs["min_interaction_pae"] <= 2.0)
    confs["min_interaction_pae<2.5"] = bool(confs["min_interaction_pae"] <= 2.5)
    confs["min_interaction_pae<3"] = bool(confs["min_interaction_pae"] <= 3)
    confs["min_interaction_pae<4"] = bool(confs["min_interaction_pae"] <= 4)
    confs["min_interaction_pae<5"] = bool(confs["min_interaction_pae"] <= 5)
    confs["design_ptm>80"] = bool(confs["design_ptm"] >= 0.8)
    confs["design_ptm>75"] = bool(confs["design_ptm"] >= 0.75)
    confs["design_iptm>80"] = bool(confs["design_iptm"] >= 0.8)
    confs["design_iptm>70"] = bool(confs["design_iptm"] >= 0.7)
    confs["design_iptm>60"] = bool(confs["design_iptm"] >= 0.6)
    confs["design_iptm>50"] = bool(confs["design_iptm"] >= 0.5)

    prefixed_metrics = {f"{prefix}{k}": v for k, v in metrics.items()}
    prefixed_metrics.update(confs)
    return prefixed_metrics


def count_noncovalents(feat):
    metrics = {}
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        biotite_array = biotite_array_from_feat(feat)
        biotite_array, _ = hydride.add_hydrogen(biotite_array)
        hbond = biotite.structure.hbond(biotite_array)
    donor_idxs, acceptor_idxs = hbond[:, 0], hbond[:, 2]
    cross_design_hbonds = (
        biotite_array.is_design[donor_idxs] != biotite_array.is_design[acceptor_idxs]
    )
    metrics["plip_hbonds"] = int(cross_design_hbonds.sum())

    # saltbridges
    pos_atoms = biotite_array[biotite_array.charge > 0]
    neg_atoms = biotite_array[biotite_array.charge < 0]
    if len(neg_atoms) > 0 and len(pos_atoms) > 0:
        pos_neg_distances = torch.cdist(
            torch.as_tensor(pos_atoms.coord), torch.as_tensor(neg_atoms.coord)
        )
        pos_idxs, neg_idxs = torch.where(
            (pos_neg_distances > 0.5) & (pos_neg_distances < 5.5)
        )
        # only keep the ones between design and non design
        cross_design_saltbridges = (
            pos_atoms.is_design[pos_idxs] != neg_atoms.is_design[neg_idxs]
        )
        metrics["plip_saltbridge"] = int(cross_design_saltbridges.sum())
    else:
        metrics["plip_saltbridge"] = 0
    return metrics


def tm_score(coords1, coords2):
    num_atoms1 = coords1.shape[0]
    num_atoms2 = coords2.shape[0]

    atom_array1 = structure.AtomArray(num_atoms1)
    atom_array1.coord = coords1.numpy()
    atom_array1.element = np.array(["C"] * num_atoms1)
    atom_array1.atom_name = np.array(["CA"] * num_atoms1)
    atom_array1.res_name = np.array(["ALA"] * num_atoms1)
    atom_array1.chain_id = np.array(["A"] * num_atoms1)
    atom_array1.res_id = np.arange(1, num_atoms1 + 1)

    atom_array2 = structure.AtomArray(num_atoms2)
    atom_array2.coord = coords2.numpy()
    atom_array2.element = np.array(["C"] * num_atoms2)
    atom_array2.atom_name = np.array(["CA"] * num_atoms2)
    atom_array2.res_name = np.array(["ALA"] * num_atoms2)
    atom_array2.chain_id = np.array(["A"] * num_atoms2)
    atom_array2.res_id = np.arange(1, num_atoms2 + 1)

    try:
        # This fails with a value error if the structures are too dissimilar. In that event, we return 0 as the TM-Score
        aligned, transform, fixed_indices, mobile_indices = (
            structure.superimpose_structural_homologs(
                atom_array1, atom_array2, max_iterations=25
            )
        )
        tm_align_fixed = structure.tm_score(
            atom_array1,
            aligned,
            fixed_indices,
            mobile_indices,
        )
    except:
        tm_align_fixed = 0

    tm_score_rmsd_aligned = 0
    if num_atoms1 == num_atoms2:
        coords1 = weighted_rigid_align(
            coords1.float()[None],
            coords2.float()[None],
            weights=torch.ones(len(coords1)).float()[None],
            mask=torch.ones(len(coords2))[None],
        ).squeeze()

        atom_array1 = structure.AtomArray(num_atoms1)
        atom_array1.coord = coords1.numpy()
        atom_array1.element = np.array(["C"] * num_atoms1)
        atom_array1.atom_name = np.array(["CA"] * num_atoms1)
        atom_array1.res_name = np.array(["ALA"] * num_atoms1)
        atom_array1.chain_id = np.array(["A"] * num_atoms1)
        atom_array1.res_id = np.arange(1, num_atoms1 + 1)

        atom_array2 = structure.AtomArray(num_atoms2)
        atom_array2.coord = coords2.numpy()
        atom_array2.element = np.array(["C"] * num_atoms2)
        atom_array2.atom_name = np.array(["CA"] * num_atoms2)
        atom_array2.res_name = np.array(["ALA"] * num_atoms2)
        atom_array2.chain_id = np.array(["A"] * num_atoms2)
        atom_array2.res_id = np.arange(1, num_atoms2 + 1)
        try:
            _, _, fixed_indices, mobile_indices = (
                structure.superimpose_structural_homologs(
                    atom_array1, atom_array2, max_iterations=25
                )
            )
            tm_score_rmsd_aligned = structure.tm_score(
                atom_array1,
                atom_array2,
                fixed_indices,
                mobile_indices,
            )
        except:
            pass

    return tm_score_rmsd_aligned, tm_align_fixed


def vendi_from_sim(mat):
    mat = mat + mat.T
    np.fill_diagonal(mat, 1.0)
    eigvals, _ = np.linalg.eigh(mat / len(mat))
    eigvals = np.clip(eigvals, 0.0, None)
    return np.exp(np.nansum(-(eigvals * np.log(eigvals))))


def vendi_scores(
    all_ca_coords: List[np.ndarray],
    all_metrics: list = None,
    fold_metrics: bool = False,
    diversity_subset: int = None,
    compute_lddts: bool = True,
    compute_iptms: bool = True,
    compute_min_int_paes: bool = True,
    backbone_fold_metrics: bool = False,
    allatom_fold_metrics: bool = True,
) -> float:
    if fold_metrics or diversity_subset is not None:
        assert all_metrics is not None
    if all_metrics is not None:
        assert len(all_ca_coords) == len(all_metrics)
    if diversity_subset is not None and diversity_subset < len(all_ca_coords):
        indices = random.sample(range(len(all_ca_coords)), diversity_subset)
        all_metrics = [all_metrics[i] for i in indices]
        all_ca_coords = [all_ca_coords[i] for i in indices]
    N = len(all_ca_coords)
    tm = np.zeros((N, N), dtype=np.float32)
    tm_fixed = np.zeros((N, N), dtype=np.float32)

    for i in tqdm(range(N), desc="Computing structure diversity."):
        for j in range(i + 1, N):
            tm_score_rmsd_aligned, tm_fixeds = tm_score(
                all_ca_coords[i], all_ca_coords[j]
            )
            tm[i, j] = tm_score_rmsd_aligned
            tm_fixed[i, j] = tm_fixeds

    scores = {
        "vendi_tm_fixed": vendi_from_sim(tm_fixed),
        "vendi_tm_align": vendi_from_sim(tm),
    }
    prefixes = []
    if allatom_fold_metrics:
        prefixes.append("")
    if backbone_fold_metrics:
        prefixes.append("bb_")
    for prefix in prefixes:
        mask_2 = np.array([m[f"{prefix}designability_rmsd_2"] for m in all_metrics])
        mask_4 = np.array([m[f"{prefix}designability_rmsd_4"] for m in all_metrics])
        mask_25 = np.array([m[f"{prefix}rmsd<2.5"] for m in all_metrics])
        mask_target_25 = np.array(
            [m[f"{prefix}target_aligned<2.5"] for m in all_metrics]
        )
        scores.update(
            {
                f"vendi_tm_{prefix}rmsd<2.5": vendi_from_sim(
                    tm_fixed[mask_25][:, mask_25]
                )
                if np.sum(mask_25) > 0
                else 0.0,
                f"vendi_tm_{prefix}rmsd_2": vendi_from_sim(tm_fixed[mask_2][:, mask_2])
                if np.sum(mask_2) > 0
                else 0.0,
                f"vendi_tm_{prefix}rmsd_4": vendi_from_sim(tm_fixed[mask_4][:, mask_4])
                if np.sum(mask_4) > 0
                else 0.0,
                f"vendi_tm_{prefix}target_aligned_rmsd<2.5": vendi_from_sim(
                    tm_fixed[mask_target_25][:, mask_target_25]
                )
                if np.sum(mask_target_25) > 0
                else 0.0,
            }
        )

        if compute_lddts:
            mask_60 = np.array(
                [m[f"{prefix}designability_lddt_60"] for m in all_metrics]
            )
            mask_65 = np.array(
                [m[f"{prefix}designability_lddt_65"] for m in all_metrics]
            )
            mask_70 = np.array(
                [m[f"{prefix}designability_lddt_70"] for m in all_metrics]
            )
            mask_75 = np.array(
                [m[f"{prefix}designability_lddt_75"] for m in all_metrics]
            )
            mask_80 = np.array(
                [m[f"{prefix}designability_lddt_80"] for m in all_metrics]
            )
            mask_85 = np.array(
                [m[f"{prefix}designability_lddt_85"] for m in all_metrics]
            )
            mask_90 = np.array(
                [m[f"{prefix}designability_lddt_90"] for m in all_metrics]
            )
            scores.update(
                {
                    f"vendi_tm_{prefix}lddt_60": vendi_from_sim(
                        tm_fixed[mask_60][:, mask_60]
                    )
                    if np.sum(mask_60) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}lddt_65": vendi_from_sim(
                        tm_fixed[mask_65][:, mask_65]
                    )
                    if np.sum(mask_65) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}lddt_70": vendi_from_sim(
                        tm_fixed[mask_70][:, mask_70]
                    )
                    if np.sum(mask_70) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}lddt_75": vendi_from_sim(
                        tm_fixed[mask_75][:, mask_75]
                    )
                    if np.sum(mask_75) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}lddt_80": vendi_from_sim(
                        tm_fixed[mask_80][:, mask_80]
                    )
                    if np.sum(mask_80) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}lddt_85": vendi_from_sim(
                        tm_fixed[mask_85][:, mask_85]
                    )
                    if np.sum(mask_85) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}lddt_90": vendi_from_sim(
                        tm_fixed[mask_90][:, mask_90]
                    )
                    if np.sum(mask_90) > 0
                    else 0.0,
                }
            )
        if compute_iptms:
            mask_80 = np.array([m[f"design_iptm>80"] for m in all_metrics])
            mask_70 = np.array([m[f"design_iptm>70"] for m in all_metrics])
            mask_60 = np.array([m[f"design_iptm>60"] for m in all_metrics])
            mask_50 = np.array([m[f"design_iptm>50"] for m in all_metrics])
            scores.update(
                {
                    f"vendi_tm_{prefix}design_iptm_80": vendi_from_sim(
                        tm_fixed[mask_80][:, mask_80]
                    )
                    if np.sum(mask_80) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}design_iptm_70": vendi_from_sim(
                        tm_fixed[mask_70][:, mask_70]
                    )
                    if np.sum(mask_70) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}design_iptm_60": vendi_from_sim(
                        tm_fixed[mask_60][:, mask_60]
                    )
                    if np.sum(mask_60) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}design_iptm_50": vendi_from_sim(
                        tm_fixed[mask_50][:, mask_50]
                    )
                    if np.sum(mask_50) > 0
                    else 0.0,
                }
            )
        if compute_min_int_paes:
            mask_15 = np.array([m[f"min_interaction_pae<1.5"] for m in all_metrics])
            mask_2 = np.array([m[f"min_interaction_pae<2"] for m in all_metrics])
            mask_25 = np.array([m[f"min_interaction_pae<2.5"] for m in all_metrics])
            mask_3 = np.array([m[f"min_interaction_pae<3"] for m in all_metrics])
            mask_4 = np.array([m[f"min_interaction_pae<4"] for m in all_metrics])
            mask_5 = np.array([m[f"min_interaction_pae<5"] for m in all_metrics])
            scores.update(
                {
                    f"vendi_tm_{prefix}min_interaction_pae_1.5": vendi_from_sim(
                        tm_fixed[mask_15][:, mask_15]
                    )
                    if np.sum(mask_15) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}min_interaction_pae_2": vendi_from_sim(
                        tm_fixed[mask_2][:, mask_2]
                    )
                    if np.sum(mask_2) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}min_interaction_pae_2.5": vendi_from_sim(
                        tm_fixed[mask_25][:, mask_25]
                    )
                    if np.sum(mask_25) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}min_interaction_pae_3": vendi_from_sim(
                        tm_fixed[mask_3][:, mask_3]
                    )
                    if np.sum(mask_3) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}min_interaction_pae_4": vendi_from_sim(
                        tm_fixed[mask_4][:, mask_4]
                    )
                    if np.sum(mask_4) > 0
                    else 0.0,
                    f"vendi_tm_{prefix}min_interaction_pae_5": vendi_from_sim(
                        tm_fixed[mask_5][:, mask_5]
                    )
                    if np.sum(mask_5) > 0
                    else 0.0,
                }
            )

    return scores


def vendi_sequences(all_seqs: List[np.ndarray], diversity_subset: int = None) -> float:
    if diversity_subset is not None and diversity_subset < len(all_seqs):
        all_seqs = random.sample(all_seqs, diversity_subset)

    N = len(all_seqs)
    sims = np.zeros((N, N), dtype=np.float32)
    aligner = Align.PairwiseAligner()
    for i in tqdm(range(N), desc="Computing sequence diversity."):
        for j in range(i + 1, N):
            seq1 = Seq(all_seqs[i])
            seq2 = Seq(all_seqs[j])
            alignments = aligner.align(seq1, seq2)

            similarity = alignments[0].score / max(len(seq1), len(seq2))
            sims[i, j] = similarity

    return {
        "vendi_seq_sim": vendi_from_sim(sims),
    }


def compute_novelty_foldseek(
    indir: Path,
    outdir: Path,
    reference_db: Path,
    files: List[str],
    foldseek_binary: str = "/data/rbg/users/hstark/foldseek/bin/foldseek",
) -> pd.DataFrame:
    if len(files) == 0:
        return np.nan

    aln_tsv = outdir / "aln.tsv"
    tmp_dir = outdir / "tmp"

    cmd = [
        foldseek_binary,
        "easy-search",
        str(indir),
        str(reference_db),
        str(aln_tsv),
        str(tmp_dir),
        "--format-output",
        "query,target,alntmscore,qtmscore,ttmscore",
        "--alignment-type",
        "1",
        "--exhaustive-search",
        "1",
    ]

    subprocess.run(cmd, check=True)

    df = pd.read_csv(
        aln_tsv,
        sep="\t",
        names=["query", "target", "alntmscore", "qtmscore", "ttmscore"],
    )
    df["tmscore"] = (df["qtmscore"] + df["ttmscore"]) / 2
    df = df.groupby("query").max().reset_index()
    queries = [Path(f).stem for f in files]
    df = df.set_index("query").reindex(queries, fill_value=0.0).reset_index()
    df_novelty = df[["query", "tmscore"]].rename(columns={"tmscore": "novelty"})
    return df_novelty


def _radius(res_name: str, atom_name: str, element: str) -> float:
    """
    ProtOr radius with element fallback.
    """
    try:
        r = vdw_radius_protor(res_name, atom_name)
        if r is not None:
            return r
    except KeyError:
        pass
    r = vdw_radius_single(element)
    return r if r is not None else 1.8


def compute_sasa(structure_path):
    HYDROPHOBIC_RESIDUES = {"ALA", "VAL", "LEU", "ILE", "MET", "PHE", "PRO", "TRP"}
    atoms = _load_stack(structure_path)[0]
    res_names = [
        bytes(r).decode() if isinstance(r, bytes) else str(r) for r in atoms.res_name
    ]
    atom_names = [
        bytes(a).decode() if isinstance(a, bytes) else str(a) for a in atoms.atom_name
    ]
    elements = [
        bytes(e).decode() if isinstance(e, bytes) else str(e) for e in atoms.element
    ]

    radii = np.array(
        [
            _radius(rn.strip(), an.strip(), el)
            for rn, an, el in zip(res_names, atom_names, elements)
        ],
        dtype=float,
    )
    atom_sasa = sasa(atoms, probe_radius=1.4, point_number=960, vdw_radii=radii)
    mask = (
        np.array([rn.strip() in HYDROPHOBIC_RESIDUES for rn in res_names])
        & np.char.startswith(atom_names, "C")
        & (atom_sasa > 0)
    )

    return atoms.coord[mask], atom_sasa[mask]


def largest_hydrophobic_patch_area(cif_path, distance_cutoff=6.0):
    result = compute_sasa(cif_path)
    if result is None or result[0].size == 0:
        return np.nan

    coords, sasa_vals = result
    if len(coords) == 0:
        return 0.0
    clustering = DBSCAN(eps=distance_cutoff, min_samples=1).fit(coords)
    labels = clustering.labels_

    max_patch_area = 0.0
    for label in np.unique(labels):
        area = sasa_vals[labels == label].sum()
        max_patch_area = max(max_patch_area, area)

    return max_patch_area


def get_delta_sasa(path, atom_design_mask):
    stack = _load_stack(path)
    atoms = stack[0]
    res = [
        r.decode().strip() if isinstance(r, bytes) else str(r).strip()
        for r in atoms.res_name
    ]
    atm = [
        a.decode().strip() if isinstance(a, bytes) else str(a).strip()
        for a in atoms.atom_name
    ]
    elem = [e.decode() if isinstance(e, bytes) else str(e) for e in atoms.element]

    radii = np.array(
        [_radius(rn, an, el) for rn, an, el in zip(res, atm, elem)], dtype=float
    )
    area = sasa(atoms, probe_radius=1.4, point_number=960, vdw_radii=radii)
    design_bound = area[atom_design_mask].sum()

    ligand_atoms = atoms[atom_design_mask]
    lig_res = [r for r, m in zip(res, atom_design_mask) if m]
    lig_atm = [a for a, m in zip(atm, atom_design_mask) if m]
    lig_elem = [e for e, m in zip(elem, atom_design_mask) if m]

    radii_lig = np.array(
        [_radius(rn, an, el) for rn, an, el in zip(lig_res, lig_atm, lig_elem)],
        dtype=float,
    )
    ligand_area = sasa(
        ligand_atoms, probe_radius=1.4, point_number=960, vdw_radii=radii_lig
    )
    delta = ligand_area.sum() - design_bound
    return delta, ligand_area.sum(), design_bound


def compute_ss_metrics(dssp_pred, ss_conditioning_metricsed):
    ss_metrics = {}
    conditioned_mask = ss_conditioning_metricsed != 0
    if conditioned_mask.sum() == 0:
        return {
            "precision_loop": float("nan"),
            "recall_loop": float("nan"),
            "accuracy_loop": float("nan"),
            "precision_helix": float("nan"),
            "recall_helix": float("nan"),
            "accuracy_helix": float("nan"),
            "precision_sheet": float("nan"),
            "recall_sheet": float("nan"),
            "accuracy_sheet": float("nan"),
            "accuracy_overall": float("nan"),
        }
    types = {1: "loop", 2: "helix", 3: "sheet"}
    TP_total, total_conditioned = 0, conditioned_mask.sum().item()
    for i, name in types.items():
        TP = ((dssp_pred == i) & (ss_conditioning_metricsed == i)).sum().item()
        FP = (
            ((dssp_pred == i) & (ss_conditioning_metricsed != i) & conditioned_mask)
            .sum()
            .item()
        )
        FN = ((dssp_pred != i) & (ss_conditioning_metricsed == i)).sum().item()
        precision = TP / (TP + FP) if (TP + FP) > 0 else float("nan")
        recall = TP / (TP + FN) if (TP + FN) > 0 else float("nan")
        accuracy = TP / (TP + FP + FN) if (TP + FP + FN) > 0 else float("nan")
        ss_metrics[f"precision_{name}"] = precision
        ss_metrics[f"recall_{name}"] = recall
        ss_metrics[f"accuracy_{name}"] = accuracy
        TP_total += TP

    accuracy_overall = (
        TP_total / total_conditioned if total_conditioned > 0 else float("nan")
    )
    ss_metrics["accuracy_overall"] = accuracy_overall
    return ss_metrics


def filter_resolved_atoms(structure: Structure) -> Structure:
    resolved_atom_indices = np.where(structure.atoms["is_present"])[0]
    return Structure.extract_atoms(structure, resolved_atom_indices)


def save_design_only_structure_to_cif(atom_design_mask, structure, output_path: Path):
    design_atom_indices = torch.where(atom_design_mask)[0].cpu().numpy()
    design_only_str = Structure.extract_atoms(
        structure, design_atom_indices, res_reindex=True
    )
    cif_text = to_mmcif(design_only_str)
    output_path.write_text(cif_text)
    return cif_text


def save_design_only_structure_to_pdb(atom_design_mask, structure, output_path: Path):
    cif_path = output_path.with_suffix(".cif")
    cif_text = save_design_only_structure_to_cif(atom_design_mask, structure, cif_path)

    cif_io = io.StringIO(cif_text)
    mmcif_parser = PDB.MMCIFParser()
    pdb_writer = PDB.PDBIO()
    parsed_structure = mmcif_parser.get_structure("des_only", cif_io)
    pdb_writer.set_structure(parsed_structure)
    pdb_writer.save(str(output_path))


########################################################################################################
# Hydrophobycity computation functions. From Jeremie Alexander.
########################################################################################################


def calc_base_h(seq: str) -> float:
    s = seq.upper()
    n = len(s)
    if n == 0:
        return 0.0
    H = 0.0

    # position-specific coefficients
    for i, aa in enumerate(s):
        if aa not in const.hydrophobicity_info:
            raise ValueError(f"Unknown residue '{aa}' in '{seq}'")
        if i == 0:
            key = "Rc1"
        elif i == 1:
            key = "Rc2"
        elif i == n - 1:
            key = "Rn"
        elif i == n - 2:
            key = "Rn1"
        else:
            key = "Rc"
        H += const.hydrophobicity_info[aa][key]

    # nearest-neighbor penalties around H/R/K
    for i, aa in enumerate(s):
        if aa in ("H", "R", "K"):
            for j in (i - 1, i + 1):
                if 0 <= j < n and s[j] in const.nn_penalty:
                    H -= const.nn_penalty[s[j]]

    # proline run penalties
    i = 0
    while i < n:
        if s[i] == "P":
            j = i
            while j < n and s[j] == "P":
                j += 1
            run = j - i
            if run >= 4:
                H -= 5.0
            elif run == 3:
                H -= 3.5
            elif run == 2:
                H -= 1.2
            i = j
        else:
            i += 1
    return H


def apply_length_weight(H: float, n: int) -> float:
    if n < 8:
        KL = 1.0 - 0.055 * (8 - n)
    elif n > 20:
        KL = 1.0 / (1.0 + 0.027 * (n - 20))
    else:
        KL = 1.0
    return H * KL


def overall_penalty(H: float) -> float:
    if H <= 20:
        return H
    if H <= 30:
        return H - 0.27 * (H - 18.0)
    if H <= 40:
        return H - 0.33 * (H - 18.0)
    if H <= 50:
        return H - 0.38 * (H - 18.0)
    return H - 0.447 * (H - 18.0)


def calc_hydrophobicity(seq: str) -> float:
    s = (seq or "").strip().upper()
    if not s or "X" in s:
        return float("nan")
    base = calc_base_h(s)
    base = apply_length_weight(base, len(s))
    return round(overall_penalty(base), 4)


########################################################################################################
# Liability computation functions
########################################################################################################


def get_motif_set(modality="antibody", peptide_type="linear"):
    """
    Returns a dict of motif_name -> compiled regex for the given modality.
    modality: 'antibody' or 'peptide'
    peptide_type: 'linear' or 'cyclic' (only for peptide modality)


    """
    if modality == "antibody":
        return {
            "DeAmdH": re.compile(r"N[GS]"),  # High-risk deamidation
            "DeAmdM": re.compile(r"N[AHNT]"),  # Medium-risk deamidation
            "DeAmdL": re.compile(r"[STK]N"),  # Low-risk deamidation
            "Ngly": re.compile(r"N[^P][ST]"),  # N-glycosylation sites
            "Isom": re.compile(r"D[DG HST]".replace(" ", "")),  # Isomerization
            "Isomer": re.compile(r"DG|DS|DD"),  # Isomerization variants
            "FragH": re.compile(r"DP"),  # High fragmentation risk
            "FragM": re.compile(r"TS"),  # Medium fragmentation risk
            "TrpOx": re.compile(r"W"),  # Tryptophan oxidation
            "MetOx": re.compile(r"M"),  # Methionine oxidation
            "Hydro": re.compile(r"NP"),  # Hydrolysis prone
            "IntBind": re.compile(r"GPR|RGD|RYD|LDV|DGE|KGD|NGR"),  # Integrin binding
            "Polyreactive": re.compile(
                r"GGG|GG|RR|VG|VVV|WWW|YY|WxW"
            ),  # Polyreactivity
            "AggPatch": re.compile(r"FHW"),  # Aggregation patches
            "ViscPatch": re.compile(r"HYF|HWH"),  # Viscosity patches
            "DeAmdH": re.compile(r"N[GS]"),
            "DeAmdM": re.compile(r"N[AHNT]"),
            "DeAmdL": re.compile(r"[STK]N"),
            "Ngly": re.compile(r"N[^P][ST]"),
            "Isom": re.compile(r"D[DG HST]".replace(" ", "")),
            "Isomer": re.compile(r"DG|DS|DD"),
            "FragH": re.compile(r"DP"),
            "FragM": re.compile(r"TS"),
            "TrpOx": re.compile(r"W"),
            "MetOx": re.compile(r"M"),
            "Hydro": re.compile(r"NP"),
            "IntBind": re.compile(r"GPR|RGD|RYD|LDV|DGE|KGD|NGR"),
            "Polyreactive": re.compile(r"GGG|GG|RR|VG|VVV|WWW|YY|WxW"),
            "AggPatch": re.compile(r"FHW"),
            "ViscPatch": re.compile(r"HYF|HWH"),
        }
    elif modality == "peptide":
        motifs = {
            "AspBridge": re.compile(r"N[GSQA]"),  # Deamidation hotspots
            "AspCleave": re.compile(r"D[PGS]"),  # Acidic cleavage sites
            "NTCycl": re.compile(r"^[QN]"),  # N-terminal cyclization
            "ProtTryp": re.compile(r"[KR](?=.)"),  # Trypsin cleavage sites
            "DPP4": re.compile(r"^[PX]?[AP]"),  # DPP4 cleavage sites
            "MetOx": re.compile(r"M"),  # Methionine oxidation
            "TrpOx": re.compile(r"W"),  # Tryptophan oxidation
            "HydroPatch": re.compile(r"[FILVWY]{3,}"),  # Hydrophobic patches
        }
        if peptide_type == "cyclic":
            # remove N-term liabilities for cyclic peptides
            motifs.pop("NTCycl", None)
            motifs.pop("DPP4", None)
        elif peptide_type == "linear":
            # For linear peptides, we handle cysteine pairing separately
            # so we don't include CysOx in the motif set
            pass
        return motifs
    else:
        raise ValueError(f"Unknown modality: {modality}")


def severity_score(name):
    return const.liability_severity.get(name, const.default_severity)


def compute_liability_scores(sequences, modality="antibody", peptide_type="linear"):
    """
    Compute liability scores for given sequences.
    modality: 'antibody' or 'peptide'; peptide_type: 'linear' or 'cyclic'.
    For cyclic peptides, terminal CysOx flags are skipped.

    Returns:
        dict: sequence -> {'score': int, 'violations': list of dicts}
    """
    motifs = get_motif_set(modality, peptide_type)
    results = {}
    for seq in sequences:
        violations = []
        total_score = 0
        length = len(seq)
        # motif scanning
        for name, pat in motifs.items():
            for m in pat.finditer(seq):
                pos = m.start() + 1
                # skip terminal cysteines for cyclic peptides
                if (
                    modality == "peptide"
                    and peptide_type == "cyclic"
                    and name == "CysOx"
                    and pos in (1, length)
                ):
                    continue
                sev = severity_score(name)
                violations.append(
                    {"motif": name, "pos": pos, "len": len(m.group()), "severity": sev}
                )
                total_score += sev
        # antibody-specific extras
        if modality == "antibody":
            # unpaired cysteines
            cpos = [i for i, aa in enumerate(seq) if aa == "C"]
            paired = set()
            for i in range(len(cpos) - 1):
                if abs(cpos[i + 1] - cpos[i]) in (1, 2):
                    paired.update({cpos[i], cpos[i + 1]})
            for i in cpos:
                if i not in paired:
                    sev = severity_score("UnpairedCys")
                    violations.append(
                        {
                            "motif": "UnpairedCys",
                            "pos": i + 1,
                            "len": 1,
                            "severity": sev,
                        }
                    )
                    total_score += sev
            # net charge
            charge = seq.count("K") + seq.count("R") - seq.count("D") - seq.count("E")
            if charge > 1:
                sev = const.default_severity
                violations.append(
                    {"motif": "HighNetCharge", "pos": None, "len": 0, "severity": sev}
                )
                total_score += sev
        # peptide-specific extras
        elif modality == "peptide":
            # For linear peptides, only flag unpaired cysteines (odd number of cysteines)
            # For cyclic peptides, terminal cysteines are expected, so only flag internal unpaired cysteines
            cpos = [i for i, aa in enumerate(seq) if aa == "C"]
            if peptide_type == "linear":
                # For linear peptides, if there's an odd number of cysteines, flag all cysteines as potential liabilities
                # since we don't know which one is unpaired
                if len(cpos) % 2 == 1:
                    sev = severity_score("UnpairedCys")
                    for cys_pos in cpos:
                        violations.append(
                            {
                                "motif": "UnpairedCys",
                                "pos": cys_pos + 1,
                                "len": 1,
                                "severity": sev,
                            }
                        )
                        total_score += sev
            elif peptide_type == "cyclic":
                # For cyclic peptides, terminal cysteines are expected for cyclization
                # Only flag internal unpaired cysteines
                internal_cpos = [
                    i for i in cpos if i != 0 and i != len(seq) - 1
                ]  # exclude terminal positions
                if len(internal_cpos) % 2 == 1:
                    # Flag all internal cysteines as potential liabilities
                    sev = severity_score("UnpairedCys")
                    for cys_pos in internal_cpos:
                        violations.append(
                            {
                                "motif": "UnpairedCys",
                                "pos": cys_pos + 1,
                                "len": 1,
                                "severity": sev,
                            }
                        )
                        total_score += sev

                # Additional liability checks for cyclic peptides
                # 1. Check for low hydrophilic content (< 40%)
                hydrophilic_residues = (
                    seq.count("D")
                    + seq.count("E")
                    + seq.count("K")
                    + seq.count("R")
                    + seq.count("H")
                    + seq.count("N")
                    + seq.count("Q")
                    + seq.count("S")
                    + seq.count("T")
                )
                hydrophilic_percentage = (hydrophilic_residues / len(seq)) * 100
                if hydrophilic_percentage < 40:
                    sev = severity_score("LowHydrophilic")
                    violations.append(
                        {
                            "motif": "LowHydrophilic",
                            "pos": None,
                            "len": 0,
                            "severity": sev,
                            "details": f"{hydrophilic_percentage:.1f}% hydrophilic",
                        }
                    )
                    total_score += sev

                # 2. Check for consecutive identical residues
                max_consec_identical = 1
                current_consec = 1
                for i in range(1, len(seq)):
                    if seq[i] == seq[i - 1]:
                        current_consec += 1
                        max_consec_identical = max(max_consec_identical, current_consec)
                    else:
                        current_consec = 1

                if max_consec_identical > 1:
                    sev = severity_score("ConsecIdentical")
                    violations.append(
                        {
                            "motif": "ConsecIdentical",
                            "pos": None,
                            "len": 0,
                            "severity": sev,
                            "details": f"{max_consec_identical} consecutive identical",
                        }
                    )
                    total_score += sev

                # 3. Check for more than 4 consecutive hydrophobic residues
                max_consec_hydrophobic = 0
                current_consec = 0
                for aa in seq:
                    if aa in "FILVWY":
                        current_consec += 1
                        max_consec_hydrophobic = max(
                            max_consec_hydrophobic, current_consec
                        )
                    else:
                        current_consec = 0

                if max_consec_hydrophobic > 4:
                    sev = severity_score("LongHydrophobic")
                    violations.append(
                        {
                            "motif": "LongHydrophobic",
                            "pos": None,
                            "len": 0,
                            "severity": sev,
                            "details": f"{max_consec_hydrophobic} consecutive hydrophobic",
                        }
                    )
                    total_score += sev
        results[seq] = {"score": total_score, "violations": violations}
    return results


def compute_liability_metrics(sequence, liability_modality, liability_peptide_type):
    metrics = {}
    # check if sequence is valid
    if not sequence or len(sequence) == 0:
        raise ValueError(f"Sequence is empty: '{sequence}'")

    liability_results = compute_liability_scores(
        [sequence],
        modality=liability_modality,
        peptide_type=liability_peptide_type,
    )
    liability_data = liability_results[sequence]

    # Store liability metrics
    metrics["liability_score"] = liability_data["score"]
    metrics["liability_num_violations"] = len(liability_data["violations"])

    # Count violations by severity
    high_severity_violations = [
        v for v in liability_data["violations"] if v["severity"] >= 10
    ]
    medium_severity_violations = [
        v for v in liability_data["violations"] if 5 <= v["severity"] < 10
    ]
    low_severity_violations = [
        v for v in liability_data["violations"] if v["severity"] < 5
    ]

    metrics["liability_high_severity_violations"] = len(high_severity_violations)
    metrics["liability_medium_severity_violations"] = len(medium_severity_violations)
    metrics["liability_low_severity_violations"] = len(low_severity_violations)

    # Count violations by type
    violation_counts = {}
    for v in liability_data["violations"]:
        motif = v["motif"]
        violation_counts[motif] = violation_counts.get(motif, 0) + 1

    # Store individual violation type counts as metrics
    for motif, count in violation_counts.items():
        metrics[f"liability_{motif}_count"] = count

    # Add detailed violation information
    # Group violations by type for intelligent reporting
    violations_by_type = {}
    for v in liability_data["violations"]:
        motif = v["motif"]
        if motif not in violations_by_type:
            violations_by_type[motif] = []
        violations_by_type[motif].append(v)

    # Initialize default values for all motifs to ensure consistent dataframe columns
    # Use the full motif set for the configured modality/peptide_type so columns are consistent
    all_motifs = set(
        get_motif_set(
            modality=liability_modality,
            peptide_type=liability_peptide_type,
        ).keys()
    )
    for motif in all_motifs:
        # Initialize all possible fields with default values
        metrics[f"liability_{motif}_count"] = 0
        metrics[
            f"liability_{motif}_position"
        ] = -1  # use -1 for no position (keeps int dtype)
        metrics[f"liability_{motif}_length"] = 0
        metrics[f"liability_{motif}_severity"] = 0
        metrics[f"liability_{motif}_details"] = ""
        metrics[f"liability_{motif}_positions"] = ""
        metrics[f"liability_{motif}_num_positions"] = 0
        metrics[f"liability_{motif}_global_details"] = ""
        metrics[f"liability_{motif}_avg_severity"] = 0.0

    # Store detailed violation information
    for motif, motif_violations in violations_by_type.items():
        if len(motif_violations) == 1:
            # Single violation - store all details
            v = motif_violations[0]
            # Ensure position is an integer; use -1 for non-positional violations
            metrics[f"liability_{motif}_position"] = (
                int(v["pos"]) if v["pos"] is not None else -1
            )
            metrics[f"liability_{motif}_length"] = v["len"]
            metrics[f"liability_{motif}_severity"] = v["severity"]
            if "details" in v:
                metrics[f"liability_{motif}_details"] = v["details"]
            else:
                metrics[f"liability_{motif}_details"] = ""
        else:
            # Multiple violations - store summary information
            positions = [v["pos"] for v in motif_violations if v["pos"] is not None]
            global_violations = [v for v in motif_violations if v["pos"] is None]

            if positions:
                # Store position range for positional violations
                metrics[f"liability_{motif}_positions"] = (
                    f"{min(positions)}-{max(positions)}"
                )
                metrics[f"liability_{motif}_num_positions"] = len(positions)
            else:
                metrics[f"liability_{motif}_positions"] = ""
                metrics[f"liability_{motif}_num_positions"] = 0

            if global_violations:
                # Store details for global violations
                details = [
                    v.get("details", "") for v in global_violations if v.get("details")
                ]
                if details:
                    metrics[f"liability_{motif}_global_details"] = "; ".join(details)
                else:
                    metrics[f"liability_{motif}_global_details"] = ""
            else:
                metrics[f"liability_{motif}_global_details"] = ""

            # Store average severity
            avg_severity = sum(v["severity"] for v in motif_violations) / len(
                motif_violations
            )
            metrics[f"liability_{motif}_avg_severity"] = round(avg_severity, 1)

    # Add a comprehensive violation summary for easy interpretation
    violation_summary = []
    for motif, motif_violations in violations_by_type.items():
        count = len(motif_violations)
        if count == 1:
            v = motif_violations[0]
            if v["pos"] is not None:
                violation_summary.append(f"{motif}(pos{v['pos']},sev{v['severity']})")
            else:
                details = v.get("details", "")
                violation_summary.append(
                    f"{motif}({details},sev{v['severity']})"
                    if details
                    else f"{motif}(sev{v['severity']})"
                )
        else:
            positions = [v["pos"] for v in motif_violations if v["pos"] is not None]
            if positions:
                violation_summary.append(
                    f"{motif}x{count}(pos{min(positions)}-{max(positions)},sev{motif_violations[0]['severity']})"
                )
            else:
                violation_summary.append(
                    f"{motif}x{count}(sev{motif_violations[0]['severity']})"
                )

    metrics["liability_violations_summary"] = (
        "; ".join(violation_summary) if violation_summary else ""
    )
    return metrics



from collections import defaultdict
import gc
import itertools
from pathlib import Path
import time
from typing import Dict, List

import wandb

import torch
import numpy as np
import torch._dynamo
from pytorch_lightning import LightningModule



from boltzgen.data import const
from boltzgen.model.validation import design
from boltzgen.task.analyze.analyze import Analyze
from boltzgen.task.predict.data_from_generated import collate
from boltzgen.task.predict.writer import AffinityWriter, FoldingWriter


class RefoldingValidator(design.DesignValidator):
    """Validation step implementation for Affinity."""

    def __init__(
        self,
        val_names: List[str],
        analyze_task: Analyze,
        folding_model_args: Dict,
        folding_args: Dict,
        folding_checkpoint: str,
        step_scale: float = 2,
        noise_scale: float = 0.75,
        ligand_plip: bool = False,
        atom14: bool = True,
        atom37: bool = False,
        backbone_only: bool = False,
        val_monomer: str = None,
        val_ligand: str = None,
        inverse_fold: bool = False,
        affinity_model_args: Dict = None,
        affinity_args: Dict = None,
        affinity_checkpoint: str = None,
    ) -> None:
        super().__init__(
            val_names=val_names,
            atom14=atom14,
            atom37=atom37,
            backbone_only=backbone_only,
            inverse_fold=inverse_fold,
        )


        self.analyze_task = analyze_task
        self.writer = None
        self.folding_model = None
        self.design_dir = None
        self.folding_model_args = folding_model_args
        self.folding_checkpoint = folding_checkpoint
        self.step_scale = step_scale
        self.noise_scale = noise_scale
        self.ligand_plip = ligand_plip
        self.writer = FoldingWriter(design_dir=None)

        # Affinity model / writer
        self.affinity_model_args = affinity_model_args or {}
        self.affinity_args = affinity_args or {}
        self.affinity_checkpoint = affinity_checkpoint
        self.affinity_model = None
        self.aff_writer = AffinityWriter(design_dir=None)

        self.folding_args = folding_args
        self.folding_args["keys_dict_out"] = const.eval_keys_confidence

        # This is the hardcoded order in the training dataset (FoldEverythingDataModule)
        idx = 1
        dataset_to_logname = {}
        if val_monomer:
            dataset_to_logname[idx] = "val_monomer"
            idx += 1
        if val_ligand:
            dataset_to_logname[idx] = "val_ligand"

        self.dataset_to_logname = dataset_to_logname
        self.all_refolding_data = defaultdict(list)
        self.all_refold_metrics = defaultdict(list)

    def run_model(
        self, model: LightningModule, batch: Dict[str, torch.Tensor], idx_dataset: int
    ) -> Dict[str, torch.Tensor]:
        """Compute the forward pass."""
        out = model(
            batch,
            recycling_steps=model.validation_args.recycling_steps,
            num_sampling_steps=model.validation_args.sampling_steps,
            diffusion_samples=model.validation_args.diffusion_samples,
            step_scale=self.step_scale,
            noise_scale=self.noise_scale,
        )

        return out

    def init_folding_model(self, model, logname):
        if self.folding_model is None:
            self.timestamp = time.time()
            print("Loading folding model for refolding validation.")
            model_module: LightningModule = type(model).load_from_checkpoint(
                checkpoint_path=self.folding_checkpoint,
                strict=False,
                map_location=model.device,
                predict_args=self.folding_args,
                **self.folding_model_args,
            )
            model_module.eval()
            self.folding_model = model_module
            model.log(
                f"{logname}/init_folding_model_dur",
                time.time() - self.timestamp,
                sync_dist=True,
            )

    def init_affinity_model(self, model, logname):
        if (self.affinity_checkpoint is None) or (self.affinity_args is None):
            return

        if self.affinity_model is None:
            self.timestamp = time.time()
            print("Loading affinity model for validation.")
            model_module: LightningModule = type(model).load_from_checkpoint(
                checkpoint_path=self.affinity_checkpoint,
                strict=False,
                predict_args=self.affinity_args,
                map_location=model.device,
                **self.affinity_model_args,
            )
            model_module.eval()
            self.affinity_model = model_module
            model.log(
                f"{logname}/init_affinity_model_dur",
                time.time() - self.timestamp,
                sync_dist=True,
            )

    def set_design_dir(self, design_dir):
        if design_dir != self.design_dir:
            # Set directories for the writer
            self.design_dir = design_dir
            self.writer.init_outdir(design_dir)
            self.aff_writer.init_outdir(design_dir)
            self.analyze_task.init_datasets(design_dir, load_dataset=True)

    def process(
        self,
        model: LightningModule,
        batch: Dict[str, torch.Tensor],
        out: Dict[str, torch.Tensor],
        idx_dataset: int,
        dataloader_idx: int,
        n_samples: int,
        batch_idx,
    ) -> None:
        logname = self.dataset_to_logname[dataloader_idx]
        # Take common step for folding and confidence
        start = time.time()
        if not self.inverse_fold:
            self.common_val_step(
                model, batch, out, idx_dataset=0, symmetry_correction=False
            )
            model.log(
                f"{logname}/common_val_step_dur", time.time() - start, sync_dist=True
            )
        feat_masked = out["feat_masked"]

        design_dir = f"{model.trainer.default_root_dir}/{logname}/epoch{model.current_epoch}_step{model.global_step}"

        start = time.time()
        success = self.design_val_step(
            model,
            batch,
            feat_masked,
            out,
            n_samples,
            batch_idx,
            generated_dir=design_dir,
            invalid_token="ALA",  # Use invalid token ALA here because sometimes all decodings are invalid and then we produce a sequence of only UNK and then the mmcif parser throws the error  ValueError: Alignment mismatch!
            logname=logname,
        )
        if not success:
            return
        model.log(f"{logname}/design_val_step_dur", time.time() - start, sync_dist=True)

        # Initialize folding model
        self.init_folding_model(model, logname)

        # Set design directory
        self.set_design_dir(design_dir)

        # Perform refolding
        start = time.time()
        samples = []
        for n in range(n_samples):
            cif_path = (
                Path(self.design_dir)
                / f"sample{n}_batch{batch_idx}_rank{model.trainer.global_rank}_{batch['id'][n]}.cif"
            )
            samples.append(
                self.analyze_task.data.predict_set.get_feat(
                    cif_path,
                    design_mask=feat_masked["design_mask"][n]
                    .cpu()
                    .numpy(),
                )
            )

        batch_gen = collate(samples)
        self.analyze_task.data.transfer_batch_to_device(
            batch_gen, device=model.device, dataloader_idx=None
        )
        model.log(
            f"{logname}/loading_data_for_folding_dur",
            time.time() - start,
            sync_dist=True,
        )

        start = time.time()
        refolded = self.folding_model.predict_step(batch_gen, batch_idx=batch_idx)

        # Affinity prediction
        self.init_affinity_model(model, logname)
        if self.affinity_model is not None:
            start_aff = time.time()
            aff_pred = self.affinity_model.predict_step(batch_gen, batch_idx=batch_idx)

            for k, v in aff_pred.items():
                if k not in refolded:
                    refolded[k] = v

            self.aff_writer.write_on_batch_end(
                trainer=None,
                pl_module=None,
                prediction=aff_pred,
                batch=batch_gen,
                batch_indices=None,
                batch_idx=batch_idx,
                dataloader_idx=None,
            )
            model.log(
                f"{logname}/affinity_dur", time.time() - start_aff, sync_dist=True
            )

        self.writer.write_on_batch_end(
            trainer=None,
            pl_module=None,
            prediction=refolded,
            batch=batch_gen,
            batch_indices=None,
            batch_idx=batch_idx,
            dataloader_idx=None,
        )

        model.log(f"{logname}/folding_dur", time.time() - start, sync_dist=True)

        basename = f"{design_dir}/sample{n}_batch{batch_idx}_rank{model.trainer.global_rank}_{batch['id'][0]}"
        sample_id = basename.split("/")[-1]

        if logname == "val_ligand" and self.ligand_plip:
            self.analyze_task.noncovalents_original = True

        self.analyze_task.compute_metrics(
            sample_id=sample_id,
            suffix=Path(f"rank{model.trainer.global_rank}"),
            design_dir=Path(design_dir),
        )

        data = np.load(self.analyze_task.metrics_dir / f"data_{sample_id}.npz")
        metrics = np.load(self.analyze_task.metrics_dir / f"metrics_{sample_id}.npz")
        data = {
            k: v.item() if v.shape == () else torch.tensor(v) for k, v in data.items()
        }
        metrics = {
            k: v.item() if v.shape == () else torch.tensor(v)
            for k, v in metrics.items()
        }

        (Path(design_dir) / const.folding_dirname / f"{sample_id}.npz").unlink(
            missing_ok=True
        )

        if metrics is not None:
            self.all_refold_metrics[logname].append(metrics)
            self.all_refolding_data[logname].append(data)

    def on_epoch_end(self, model):
        # Cleanup
        del self.folding_model
        self.folding_model = None
        del self.affinity_model
        self.affinity_model = None
        torch._C._cuda_clearCublasWorkspaces()
        torch._dynamo.reset()
        gc.collect()
        torch.cuda.empty_cache()

        # Compute standard metrics
        self.common_on_epoch_end(model, logname="val_monomer_ligand")
        self.on_epoch_end_design(model, logname="val_monomer_ligand")
        
        # Compute the refolding metrics
        for logname in self.dataset_to_logname.values():
            self.on_epoch_end_refolding(model, logname=logname)

        model.log(
            f"val_monomer_ligand/dur", time.time() - self.timestamp, sync_dist=True
        )

    def gather_lists(self, model: LightningModule, object):
        if torch.distributed.is_available() and torch.distributed.is_initialized():
            object_list = [None] * model.trainer.world_size
            torch.distributed.all_gather_object(object_list, object)
            return list(itertools.chain(*object_list))
        else:
            return object

    def on_epoch_end_refolding(self, model: LightningModule, logname: str):
        # Run evaluations
        start = time.time()
        design_dir = Path(
            f"{model.trainer.default_root_dir}/{logname}/epoch{model.current_epoch}_step{model.global_step}"
        )
        design_dir.mkdir(exist_ok=True, parents=True)
        design_dir = str(design_dir)
        self.analyze_task.init_datasets(design_dir)

        all_metrics = self.gather_lists(model, self.all_refold_metrics[logname])
        df, histograms = self.analyze_task.make_histograms(all_metrics)
        avg_metrics = df.mean(numeric_only=True).round(5).to_dict()
        avg_metrics["num_targets"] = len(all_metrics)

        print(f"Computing diversity for {logname}.")
        refolding_data = self.gather_lists(model, self.all_refolding_data[logname])
        diversity_metrics, _ = self.analyze_task.compute_diversity(
            refolding_data, all_metrics
        )
        for k in diversity_metrics:
            avg_metrics[k] = diversity_metrics[k]

        print(f"computing novelty {logname}.")
        novelty_metrics, _ = self.analyze_task.compute_novelty(
            suffix=Path(f"rank{model.trainer.global_rank}"),
        )
        for k in novelty_metrics:
            avg_metrics[k] = novelty_metrics[k]

        # Log results
        for name, fig in histograms.items():
            if model.logger is not None:
                model.logger.log_image(f"{logname}/{name}", [wandb.Image(fig)])

        for k, v in avg_metrics.items():
            model.log(f"{logname}/{k}", v, prog_bar=False, sync_dist=True)

        self.all_refold_metrics[logname] = []
        self.all_refolding_data[logname] = []
        model.log(
            f"{logname}/one_epoch_end_refolding_dur",
            time.time() - start,
            sync_dist=True,
        )



from typing import Dict, Tuple

import torch
from torch import Tensor


def distogram_loss(
    output: Dict[str, Tensor],
    feats: Dict[str, Tensor],
    aggregate_distogram: bool = True,
) -> Tuple[Tensor, Tensor]:
    """Compute the  distogram loss.

    Parameters
    ----------
    output : Dict[str, Tensor]
        Output of the model
    feats : Dict[str, Tensor]
        Input features

    Returns
    -------
    Tensor
        The globally averaged loss.
    Tensor
        Per example loss.

    """
    with torch.autocast("cuda", enabled=False):
        # Get predicted distograms
        pred = output["pdistogram"].float()  # (B, L, L, num_distograms, disto_bins)
        D = pred.shape[3] # num_distograms  # noqa: N806
        assert len(pred.shape) == 5  # noqa: PLR2004

        # Compute target distogram
        target = feats["disto_target"]  # (B, L, L, K, disto_bins)
        assert len(target.shape) == 5  # noqa: PLR2004

        if aggregate_distogram:
            msg = "Cannot aggregate GT distogram when num_distograms > 1"
            assert pred.shape[3] == 1, msg

            pred = pred.squeeze(3) # (B, L, L, disto_bins)

            # Aggregate distogram over K conformers
            target = target.sum(dim=3) # (B, L, L, disto_bins)

            # Normalize distogram
            P = target / target.sum(-1)[..., None].clamp(min=1)  # noqa: N806

            # Combine target mask and padding mask
            mask = feats["token_disto_mask"]
            mask = mask[:, None, :] * mask[:, :, None]
            mask = mask * (1 - torch.eye(mask.shape[1])[None]).to(pred)

            # Compute the distogram loss
            log_Q = torch.nn.functional.log_softmax(pred, dim=-1)  # noqa: N806
            errors = -1 * torch.sum(
                P * log_Q,
                dim=-1,
            )
            denom = 1e-5 + torch.sum(mask, dim=(-1, -2))
            mean = errors * mask
            mean = torch.sum(mean, dim=-1)
            mean = mean / denom[..., None]
            batch_loss = torch.sum(mean, dim=-1)
            global_loss = torch.mean(batch_loss)
        else:
            # We want to compute the loss for each pair of conformer K and predicted
            # distogram

            # Loop through conformers and compute the loss
            batch_loss = []
            for k in range(target.shape[3]):
                # Get the target distogram for conformer k
                # (B, L, L, K, disto_bins) -> (B, L, L, D, disto_bins)
                P_k = target[:, :, :, k:k+1, :].repeat_interleave(D, dim=3)  # noqa: N806

                # Compute the distogram loss to all predicted distograms
                log_Q = torch.nn.functional.log_softmax(pred, dim=-1)  # noqa: N806
                errors = -1 * torch.sum(
                    P_k * log_Q,
                    dim=-1,
                )  # (B, L, L, D)

                # Compute mask
                mask = feats["token_disto_mask"]
                mask = mask[:, None, :] * mask[:, :, None]
                mask = mask * (1 - torch.eye(mask.shape[1])[None]).to(pred)
                mask = mask.unsqueeze(-1).repeat_interleave(D, -1)  # (B, L, L, D)

                denom = 1e-5 + torch.sum(mask, dim=(-2, -3))  # (B, D)
                mean = errors * mask
                mean = torch.sum(mean, dim=-2) # (B, L, D)
                mean = mean / denom[..., None, :]
                b_loss = torch.sum(mean, dim=-2)  # (B, D)

                batch_loss.append(b_loss)

            batch_loss = torch.stack(batch_loss, dim=1)  # (B, K, D)

            # Compute the batch loss by taking the min over the predicted distograms
            # and the average across conformers
            batch_loss = torch.min(batch_loss, dim=-1).values.mean(dim=1)
            global_loss = torch.mean(batch_loss)

        return global_loss, batch_loss





from typing import List, Tuple

import torch
from torch import Tensor
from torch.nn.functional import pad


def pad_dim(data: Tensor, dim: int, pad_len: float, value: float = 0) -> Tensor:
    """Pad a tensor along a given dimension.

    Parameters
    ----------
    data : Tensor
        The input tensor.
    dim : int
        The dimension to pad.
    pad_len : float
        The padding length.
    value : int, optional
        The value to pad with.

    Returns
    -------
    Tensor
        The padded tensor.

    """
    if pad_len == 0:
        return data

    total_dims = len(data.shape)
    padding = [0] * (2 * (total_dims - dim))
    padding[2 * (total_dims - 1 - dim) + 1] = pad_len
    return pad(data, tuple(padding), value=value)


def pad_to_max(data: List[Tensor], value: float = 0) -> Tuple[Tensor, Tensor]:
    """Pad the data in all dimensions to the maximum found.

    Parameters
    ----------
    data : List[Tensor]
        List of tensors to pad.
    value : float
        The value to use for padding.

    Returns
    -------
    Tensor
        The padded tensor.
    Tensor
        The padding mask.

    """
    if isinstance(data[0], str):
        return data, 0

    # Check if all have the same shape
    if all(d.shape == data[0].shape for d in data):
        return torch.stack(data, dim=0), 0

    # Get the maximum in each dimension
    num_dims = len(data[0].shape)
    max_dims = [max(d.shape[i] for d in data) for i in range(num_dims)]

    # Get the padding lengths
    pad_lengths = []
    for d in data:
        dims = []
        for i in range(num_dims):
            dims.append(0)
            dims.append(max_dims[num_dims - i - 1] - d.shape[num_dims - i - 1])
        pad_lengths.append(dims)

    data = [pad(d, pad_len, value=value) for d, pad_len in zip(data, pad_lengths)]

    # Stack the data
    data = torch.stack(data, dim=0)

    return data



# Copyright 2021 AlQuraishi Laboratory
# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial, partialmethod
from typing import Optional

import torch
import torch.nn as nn

from boltzgen.model.layers.triangular_attention.primitives import (
    Attention,
    LayerNorm,
    Linear,
)
from boltzgen.model.layers.triangular_attention.utils import (
    chunk_layer,
    permute_final_dims,
)


class TriangleAttention(nn.Module):
    """Implement Algorithm 12."""

    def __init__(
        self,
        c_in: int,
        c_hidden: int,
        no_heads: int,
        starting: bool = True,
        inf: float = 1e9,
    ) -> None:
        super().__init__()

        self.c_in = c_in
        self.c_hidden = c_hidden
        self.no_heads = no_heads
        self.starting = starting
        self.inf = inf

        self.layer_norm = LayerNorm(self.c_in)

        self.linear = Linear(c_in, self.no_heads, bias=False, init="normal")

        self.mha = Attention(
            self.c_in, self.c_in, self.c_in, self.c_hidden, self.no_heads
        )

    @torch.jit.ignore
    def _chunk(
        self,
        x: torch.Tensor,
        tri_bias: torch.Tensor,
        mask_bias: torch.Tensor,
        mask: torch.Tensor,
        chunk_size: int,
        use_kernels: bool = False,
    ) -> torch.Tensor:
        """Compute triangle attention.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape [*, I, J, C_in]
        biases : list[torch.Tensor]
            List of bias tensors of shape [*, H, I, J]
        chunk_size : int
            Size of chunks for memory efficient computation
        use_kernels : bool, default=False
            Whether to use optimized CUDA kernels

        Returns
        -------
        torch.Tensor
            Output tensor of shape [*, I, J, C_in]

        """
        mha_inputs = {
            "q_x": x,
            "kv_x": x,
            "tri_bias": tri_bias,
            "mask_bias": mask_bias,
            "mask": mask,
        }

        return chunk_layer(
            partial(
                self.mha,
                use_kernels=use_kernels,
            ),
            mha_inputs,
            chunk_size=chunk_size,
            no_batch_dims=len(x.shape[:-2]),
            _out=None,
        )

    def forward(
        self,
        x: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
        chunk_size: Optional[int] = None,
        use_kernels: bool = False,
    ) -> torch.Tensor:
        """Compute triangle attention.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor of shape [*, I, J, C_in]
        mask : torch.Tensor, optional
            Attention mask of shape [*, I, J]
        chunk_size : int, optional
            Size of chunks for memory efficient computation
        use_kernels : bool, default=False
            Whether to use optimized CUDA kernels

        Returns
        -------
        torch.Tensor
            Output tensor of shape [*, I, J, C_in]

        """
        if mask is None:
            # [*, I, J]
            mask = x.new_ones(
                x.shape[:-1],
            )

        if not self.starting:
            x = x.transpose(-2, -3)
            mask = mask.transpose(-1, -2)

        # [*, I, J, C_in]
        x = self.layer_norm(x)

        # [*, I, 1, 1, J]
        mask = mask[..., :, None, None, :]
        mask_bias = self.inf * (mask - 1)

        # [*, H, I, J]
        triangle_bias = permute_final_dims(self.linear(x), (2, 0, 1))

        # [*, 1, H, I, J]
        triangle_bias = triangle_bias.unsqueeze(-4)

        if chunk_size is not None and not use_kernels:
            x = self._chunk(
                x,
                triangle_bias,
                mask_bias,
                mask,
                chunk_size,
                use_kernels=use_kernels,
            )
        else:
            x = self.mha(
                x,
                x,
                triangle_bias,
                mask_bias,
                mask,
                use_kernels=use_kernels,
            )

        if not self.starting:
            x = x.transpose(-2, -3)

        return x


# Implements Algorithm 13
TriangleAttentionStartingNode = TriangleAttention


class TriangleAttentionEndingNode(TriangleAttention):
    """Implement Algorithm 14."""

    __init__ = partialmethod(TriangleAttention.__init__, starting=False)



from collections import defaultdict
from typing import Dict, List, Optional, Tuple

import torch
from pytorch_lightning import LightningModule
from torch import nn
from torchmetrics import MeanMetric

from boltzgen.data import const
from boltzgen.data.rmsd_computation import get_true_coordinates
from boltzgen.model.loss.distogram import distogram_loss

from boltzgen.model.loss.validation import (
    factored_lddt_loss,
    factored_token_lddt_dist_loss,
)


class Validator(nn.Module):
    """Compute validation step and aggregation."""

    def __init__(
        self,
        val_names: List[str],
        confidence_prediction: bool = False,
        override_val_method: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.val_names = val_names

        self.override_val_method = override_val_method
        if override_val_method is not None:
            override_val_method = override_val_method.lower()
            assert override_val_method in const.method_types_ids, "Invalid method type."
            self.override_val_method = const.method_types_ids[override_val_method]

        self.num_val_datasets = num_val_datasets = len(val_names)

        msg = "Only one dataset supported for now per validator. Define multiple validators for multiple datasets."
        assert num_val_datasets == 1, msg

        ### Initialize metrics ##

        # Folding metrics
        folding_metric_labels = [
            "lddt",
            "disto_lddt",
            "complex_lddt",
            "disto_loss",
        ]
        self.fold_metrics = nn.ModuleDict(
            {
                k: nn.ModuleList([nn.ModuleDict() for _ in range(num_val_datasets)])
                for k in folding_metric_labels
            }
        )

        # Confidence metrics
        confidence_metric_prefixes = [
            "top1",
            "iplddt_top1",
            "ipde_top1",
            "pde_top1",
            "ptm_top1",
            "iptm_top1",
            "ligand_iptm_top1",
            "protein_iptm_top1",
            "avg",
        ]
        mae_metric_labels = ["plddt_mae", "pde_mae", "pae_mae"]
        lddt_confidence_metric_labels = [
            prefix + "_lddt" for prefix in confidence_metric_prefixes
        ]

        if confidence_prediction:
            self.confidence_metrics = nn.ModuleDict(
                {
                    k: nn.ModuleList([nn.ModuleDict() for _ in range(num_val_datasets)])
                    for k in lddt_confidence_metric_labels + mae_metric_labels
                }
            )

        # Initialize metrics for datasets
        for val_idx in range(num_val_datasets):
            for m_ in [
                *const.out_types,
            ]:
                self.fold_metrics["disto_lddt"][val_idx][m_] = MeanMetric()
                for suffix in [":recall", ":precision", ":diversity"]:
                    m = m_ + suffix
                    self.fold_metrics["lddt"][val_idx][m] = MeanMetric()
                    self.fold_metrics["complex_lddt"][val_idx][m] = MeanMetric()
                    if confidence_prediction:
                        for k in lddt_confidence_metric_labels + mae_metric_labels:
                            self.confidence_metrics[k][val_idx][m_] = MeanMetric()

            for m in const.out_single_types:
                if confidence_prediction:
                    self.confidence_metrics["plddt_mae"][val_idx][m] = MeanMetric()

            for m in ["disto_loss"]:
                self.fold_metrics["disto_loss"][val_idx][m] = MeanMetric()

    def run_model(
        self, model: LightningModule, batch: Dict[str, torch.Tensor], idx_dataset: int
    ) -> Dict[str, torch.Tensor]:
        """Compute the forward pass."""
        if self.override_val_method is not None:
            new_feature = batch["method_feature"] * 0 + self.override_val_method
            batch["method_feature"] = new_feature

        out = model(
            batch,
            recycling_steps=model.validation_args.recycling_steps,
            num_sampling_steps=model.validation_args.sampling_steps,
            diffusion_samples=model.validation_args.diffusion_samples,
            run_confidence_sequentially=model.validation_args.get(
                "run_confidence_sequentially", False
            ),
        )

        return out

    # @abstractmethod
    def process(
        self,
        model: LightningModule,
        batch: Dict[str, torch.Tensor],
        out: Dict[str, torch.Tensor],
        idx_dataset: int,
        dataloader_idx: int,
        n_samples: int,
    ) -> None:
        """Run a validation step.

        Parameters
        ----------
        model : LightningModule
            The LightningModule model.
        batch : Dict[str, torch.Tensor]
            The batch input.
        out : Dict[str, torch.Tensor]
            The output of the model.

        """
        raise NotImplementedError

    def get_local_val_index(self, model: LightningModule, idx_dataset: int) -> int:
        """Get the local validation index.

        Parameters
        ----------
        idx_dataset : int
            The dataset index.

        Returns
        -------
        int
            The local validation index.
        """
        val_name = model.val_group_mapper[idx_dataset]["label"]
        return self.val_names.index(val_name)

    def compute_disto_loss(
        self,
        model: LightningModule,
        out: Dict[str, torch.Tensor],
        batch: Dict[str, torch.Tensor],
        idx_dataset: int,
    ) -> None:
        """Compute distogram loss."""
        # Compute validation disto loss
        val_disto_loss, _ = distogram_loss(
            out, batch, aggregate_distogram=model.aggregate_distogram
        )

        return val_disto_loss

    def compute_disto_lddt(self, model, batch, out, idx_dataset) -> Tuple[Dict, Dict]:
        """Compute distogram lddt."""
        boundaries = torch.linspace(model.min_dist, model.max_dist, model.num_bins - 1)
        lower = torch.tensor([1.0])
        upper = torch.tensor([model.max_dist + 5.0])
        exp_boundaries = torch.cat((lower, boundaries, upper))
        mid_points = ((exp_boundaries[:-1] + exp_boundaries[1:]) / 2).to(
            out["pdistogram"]
        )

        # Compute true distogram
        K = batch["coords"].shape[1]
        true_center = batch["disto_coords"].reshape(K, -1, 3)  # (K, L, 3)

        batch["token_disto_mask"] = batch["token_disto_mask"]

        # Compute distogram lddt by looping over predicted distograms
        disto_lddt_dict = defaultdict(lambda: torch.zeros(K, 1).to(model.device))
        disto_total_dict = defaultdict(lambda: torch.zeros(K, 1).to(model.device))

        # Compute predicted dists
        preds = out["pdistogram"][:, :, :, 0]
        pred_softmax = torch.softmax(preds, dim=-1)
        pred_softmax = pred_softmax.argmax(dim=-1)
        pred_softmax = torch.nn.functional.one_hot(
            pred_softmax, num_classes=preds.shape[-1]
        )
        pred_dist_i = (pred_softmax * mid_points).sum(dim=-1)  # (B, L, L)
        del pred_softmax

        # Compute true distances for each conformer
        # Implemented in a loop to avoid memory issues with large number of
        # conformers. Batched version over K factored_token_lddt_dist_loss_ensemble
        # more efficient for small K.
        for k in range(K):
            true_dists_k = torch.cdist(true_center[k], true_center[k])[
                None
            ]  # (1, L * L)

            # Compute lddt
            disto_lddt_dict_, disto_total_dict_ = factored_token_lddt_dist_loss(
                feats=batch,
                true_d=true_dists_k,
                pred_d=pred_dist_i,
            )

        for key in disto_lddt_dict_:
            disto_lddt_dict[key][k, 0] = disto_lddt_dict_[key].item()
            disto_total_dict[key][k, 0] = disto_total_dict_[key].item()

        for key in disto_lddt_dict:
            # Take min over distograms and average over conformers. Add batch dimension.
            disto_lddt_dict[key] = (
                disto_lddt_dict[key].min(dim=1).values.mean(dim=0)[None]
            )
            disto_total_dict[key] = (
                disto_total_dict[key].min(dim=1).values.mean(dim=0)[None]
            )
        del true_center
        del preds

        return disto_lddt_dict, disto_total_dict

    def get_lddt_metrics(
        self,
        model,
        batch,
        out,
        idx_dataset,
        n_samples,
        true_coords_resolved_mask,
        true_coords,
        expand_to_diffusion_samples,
    ) -> Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]:
        K = batch["coords"].shape[1]

        if not expand_to_diffusion_samples:
            true_coords_resolved_mask = true_coords_resolved_mask.unsqueeze(0).repeat(
                (n_samples, 1)
            )

        ### Compute lddt ###
        # Implemented in a loop to avoid memory issues with large number
        # of conformers
        all_lddt_dict = defaultdict(list)
        all_total_dict = defaultdict(list)
        for ensemble_idx in range(K):
            # This OOM for large n_samples. Need to chunk or loop over samples.

            if expand_to_diffusion_samples:
                true_coords_k = true_coords[:, ensemble_idx]
            else:
                true_coords_k = (
                    true_coords[ensemble_idx].unsqueeze(0).repeat((n_samples, 1, 1))
                )

            all_lddt_dict_s, all_total_dict_s = factored_lddt_loss(
                feats=batch,
                atom_mask=true_coords_resolved_mask,
                true_atom_coords=true_coords_k,  # (multiplicity, L, 3)
                pred_atom_coords=out["sample_atom_coords"],
                multiplicity=n_samples,
                exclude_ions=model.exclude_ions_from_lddt,
            )
            for key in all_lddt_dict_s:
                all_lddt_dict[key].append(all_lddt_dict_s[key])
                all_total_dict[key].append(all_total_dict_s[key])

        for key in all_lddt_dict:
            all_lddt_dict[key] = torch.stack(
                all_lddt_dict[key], dim=1
            )  # (multiplicity, K)
            all_total_dict[key] = torch.stack(all_total_dict[key], dim=1)
        return all_lddt_dict, all_total_dict

    def compute_best_lddt_metrics(
        self,
        model,
        batch,
        all_lddt_dict,
        all_total_dict,
        n_samples,
    ):
        K = batch["coords"].shape[1]

        # if the multiplicity used is > 1 then we take the best lddt of the different samples
        # AF3 combines this with the confidence based filtering
        best_lddt_dict, best_total_dict = {}, {}
        best_complex_lddt_dict, best_complex_total_dict = {}, {}
        # B = true_coords.shape[0] // n_samples
        if n_samples > 1:
            # NOTE: we can change the way we aggregate the lddt
            complex_total = 0
            complex_lddt = 0
            for key in all_lddt_dict:
                complex_lddt += all_lddt_dict[key] * all_total_dict[key]
                complex_total += all_total_dict[key]
            complex_lddt /= complex_total + 1e-7

            # Take best over samples, average over conformers: recall groundtruth
            # conformers
            suffix = ":recall"
            best_complex_idx = complex_lddt.argmax(dim=0)
            for key in all_lddt_dict:
                # take best across diffusion samples
                best_idx = all_lddt_dict[key].argmax(dim=0)
                best_lddt_dict[key + suffix] = all_lddt_dict[key][
                    best_idx, torch.arange(K)
                ].mean(dim=0)[None]  # take average across conformers in ensemble
                best_total_dict[key + suffix] = all_total_dict[key][
                    best_idx, torch.arange(K)
                ].mean(dim=0)[None]
                # mean(dim=0) since samples was argmaxed, add back batch dim
                best_complex_lddt_dict[key + suffix] = all_lddt_dict[key][
                    best_complex_idx, torch.arange(K)
                ].mean(dim=0)[None]
                best_complex_total_dict[key + suffix] = all_total_dict[key][
                    best_complex_idx, torch.arange(K)
                ].mean(dim=0)[None]  # sum here ?

            # Take best over conformers, average over samples: precision
            suffix = ":precision"
            best_complex_idx = complex_lddt.argmax(dim=1)
            for key in all_lddt_dict:
                # take best across diffusion conformers
                best_idx = all_lddt_dict[key].argmax(dim=1)
                best_lddt_dict[key + suffix] = all_lddt_dict[key][
                    torch.arange(n_samples), best_idx
                ].mean(dim=0)[None]  # take average across samples in ensemble
                best_total_dict[key + suffix] = all_total_dict[key][
                    torch.arange(n_samples), best_idx
                ].mean(dim=0)[None]
                # dim 0 since samples was argmaxed, add back batch dim [None]
                best_complex_lddt_dict[key + suffix] = all_lddt_dict[key][
                    torch.arange(n_samples), best_complex_idx
                ].mean(dim=0)[None]
                best_complex_total_dict[key + suffix] = all_total_dict[key][
                    torch.arange(n_samples), best_complex_idx
                ].mean(dim=0)[None]

        else:
            # Take average across conformers in ensemble.
            for key in all_lddt_dict:
                for suffix in [":recall", ":precision"]:
                    best_lddt_dict[key + suffix] = (
                        all_lddt_dict[key].max(dim=1).values[None]
                    )  # (sample, K) -> (B=1, samples=1)
                    best_total_dict[key + suffix] = (
                        all_total_dict[key].max(dim=1).values[None]
                    )
                    best_complex_lddt_dict[key + suffix] = (
                        all_lddt_dict[key].max(dim=1).values[None]
                    )
                    best_complex_total_dict[key + suffix] = (
                        all_total_dict[key].max(dim=1).values[None]
                    )
                suffix = ":diversity"
                best_lddt_dict[key + suffix] = torch.tensor([0.0]).to(model.device)
                best_total_dict[key + suffix] = torch.tensor([1.0]).to(model.device)
                best_complex_lddt_dict[key + suffix] = torch.tensor([0.0]).to(
                    model.device
                )
                best_complex_total_dict[key + suffix] = torch.tensor([1.0]).to(
                    model.device
                )

        return (
            best_lddt_dict,
            best_total_dict,
            best_complex_lddt_dict,
            best_complex_total_dict,
        )

    def update_lddt_rmsd_metrics(
        self,
        batch,
        disto_lddt_dict,
        disto_total_dict,
        best_lddt_dict,
        best_total_dict,
        best_complex_lddt_dict,
        best_complex_total_dict,
        idx_dataset,
        return_dict,
    ):
        # Folding metrics
        for m_ in const.out_types:
            if m_ == "ligand_protein":
                self.fold_metrics["disto_lddt"][idx_dataset]["ligand_protein"].update(
                    disto_lddt_dict[m_], disto_total_dict[m_]
                )
                for suffix in [":recall", ":precision", ":diversity"]:
                    m = m_ + suffix
                    self.fold_metrics["lddt"][idx_dataset][
                        "ligand_protein" + suffix
                    ].update(best_lddt_dict[m], best_total_dict[m])
                    self.fold_metrics["complex_lddt"][idx_dataset][
                        "ligand_protein" + suffix
                    ].update(best_complex_lddt_dict[m], best_complex_total_dict[m])

            elif m_ == "protein_protein":
                self.fold_metrics["disto_lddt"][idx_dataset]["protein_protein"].update(
                    disto_lddt_dict[m_], disto_total_dict[m_]
                )
                for suffix in [":recall", ":precision", ":diversity"]:
                    m = m_ + suffix
                    self.fold_metrics["lddt"][idx_dataset][
                        "protein_protein" + suffix
                    ].update(best_lddt_dict[m], best_total_dict[m])
                    self.fold_metrics["complex_lddt"][idx_dataset][
                        "protein_protein" + suffix
                    ].update(best_complex_lddt_dict[m], best_complex_total_dict[m])

            else:
                self.fold_metrics["disto_lddt"][idx_dataset][m_].update(
                    disto_lddt_dict[m_], disto_total_dict[m_]
                )
                for suffix in [":recall", ":precision", ":diversity"]:
                    m = m_ + suffix
                    self.fold_metrics["lddt"][idx_dataset][m].update(
                        best_lddt_dict[m], best_total_dict[m]
                    )
                    self.fold_metrics["complex_lddt"][idx_dataset][m].update(
                        best_complex_lddt_dict[m], best_complex_total_dict[m]
                    )

    def common_val_step(
        self,
        model: LightningModule,
        batch: Dict[str, torch.Tensor],
        out: Dict[str, torch.Tensor],
        idx_dataset: int,
        expand_to_diffusion_samples: bool = True,
        symmetry_correction: bool = None,
    ) -> None:
        """Run a common validation step.

        Parameters
        ----------
        model : LightningModule
            The LightningModule model.
        batch : Dict[str, torch.Tensor]
            The batch input.
        out : Dict[str, torch.Tensor]
            The output of the model.
        """
        if symmetry_correction is None:
            symmetry_correction = model.val_group_mapper[idx_dataset][
                "symmetry_correction"
            ]  # global val index

        # Get the local validation index from the global index
        idx_dataset = self.get_local_val_index(model, idx_dataset)

        n_samples = model.validation_args.diffusion_samples

        # Compute distogram loss and update metrics

        val_disto_loss = self.compute_disto_loss(model, out, batch, idx_dataset)

        # Compute distogram lddt and update metrics

        disto_lddt_dict, disto_total_dict = self.compute_disto_lddt(
            model, batch, out, idx_dataset
        )

        # Get true coords

        return_dict = get_true_coordinates(
            batch,
            out,
            n_samples,
            symmetry_correction,
            expand_to_diffusion_samples=expand_to_diffusion_samples,
        )

        # Move this and do better as to when to interleave
        true_coords = return_dict[
            "true_coords"
        ]  # (multiplicity, K, L, 3) if expand_to_diffusion_samples else (K, L, 3)
        true_coords_resolved_mask = return_dict[
            "true_coords_resolved_mask"
        ]  # (multiplicity, L) if expand_to_diffusion_samples else (L)

        # Get lddt metrics

        all_lddt_dict, all_total_dict = self.get_lddt_metrics(
            model,
            batch,
            out,
            idx_dataset,
            n_samples,
            true_coords_resolved_mask,
            true_coords,
            expand_to_diffusion_samples,
        )

        # Compute best lddt metrics based on oracle lddt and
        # average across conformers
        (
            best_lddt_dict,
            best_total_dict,
            best_complex_lddt_dict,
            best_complex_total_dict,
        ) = self.compute_best_lddt_metrics(
            model,
            batch,
            all_lddt_dict,
            all_total_dict,
            n_samples,
        )

        ### Update the metrics ###

        # Update distogram loss
        self.fold_metrics["disto_loss"][idx_dataset]["disto_loss"].update(
            val_disto_loss
        )

        # Update folding metrics
        self.update_lddt_rmsd_metrics(
            batch,
            disto_lddt_dict,
            disto_total_dict,
            best_lddt_dict,
            best_total_dict,
            best_complex_lddt_dict,
            best_complex_total_dict,
            # rmsds,
            idx_dataset,
            return_dict,
        )

    def on_epoch_end(self, model: LightningModule):
        raise NotImplementedError

    def common_on_epoch_end(self, model: LightningModule, logname: str = "val"):
        avg_lddt = [{} for _ in range(self.num_val_datasets)]
        avg_disto_lddt = [{} for _ in range(self.num_val_datasets)]

        for idx_dataset in range(self.num_val_datasets):
            dataset_name_ori = self.val_names[idx_dataset]

            # this is harcodeded for now to compare with Boltz-1 metrics
            dataset_name = (
                ""
                if dataset_name_ori == "RCSB" or dataset_name_ori == "monomer"
                else f"__{dataset_name_ori}"
            )

            for m_ in [
                *const.out_types,
            ]:
                avg_disto_lddt[idx_dataset][m_] = self.fold_metrics["disto_lddt"][
                    idx_dataset
                ][m_].compute()

                avg_disto_lddt[idx_dataset][m_] = (
                    0.0
                    if torch.isnan(avg_disto_lddt[idx_dataset][m_])
                    else avg_disto_lddt[idx_dataset][m_].item()
                )
                self.fold_metrics["disto_lddt"][idx_dataset][m_].reset()
                model.log(
                    f"{logname}/disto_lddt_{m_}{dataset_name}",
                    avg_disto_lddt[idx_dataset][m_],
                )

                m = m_ + ":recall"
                avg_lddt[idx_dataset][m] = self.fold_metrics["lddt"][idx_dataset][
                    m
                ].compute()
                avg_lddt[idx_dataset][m] = (
                    0.0
                    if torch.isnan(avg_lddt[idx_dataset][m])
                    else avg_lddt[idx_dataset][m].item()
                )
                self.fold_metrics["lddt"][idx_dataset][m].reset()
                model.log(
                    f"{logname}/lddt_{m_}{dataset_name}",
                    avg_lddt[idx_dataset][m],
                )

            overall_disto_lddt = sum(
                avg_disto_lddt[idx_dataset][m] * w
                for (m, w) in const.out_types_weights.items()
            ) / sum(const.out_types_weights.values())
            model.log(
                f"{logname}/disto_lddt{dataset_name}",
                overall_disto_lddt,
            )

            overall_lddt = sum(
                avg_lddt[idx_dataset][m + ":recall"] * w
                for (m, w) in const.out_types_weights.items()
            ) / sum(const.out_types_weights.values())

            model.log(
                f"{logname}/lddt",
                overall_lddt,
            )

            # Distogram loss
            r = self.fold_metrics["disto_loss"][idx_dataset]["disto_loss"].compute()
            model.log(f"{logname}/disto_loss{dataset_name}", r)
            self.fold_metrics["disto_loss"][idx_dataset]["disto_loss"].reset()



import torch
from torch import nn
from boltzgen.data import const


def compute_collinear_mask(v1, v2):
    norm1 = torch.norm(v1, dim=1, keepdim=True)
    norm2 = torch.norm(v2, dim=1, keepdim=True)
    v1 = v1 / (norm1 + 1e-6)
    v2 = v2 / (norm2 + 1e-6)
    mask_angle = torch.abs(torch.sum(v1 * v2, dim=1)) < 0.9063
    mask_overlap1 = norm1.reshape(-1) > 1e-2
    mask_overlap2 = norm2.reshape(-1) > 1e-2
    return mask_angle & mask_overlap1 & mask_overlap2


def compute_frame_pred(
    pred_atom_coords,
    frames_idx_true,
    feats,
    multiplicity,
    resolved_mask=None,
    inference=False,
):
    with torch.amp.autocast("cuda", enabled=False):
        asym_id_token = feats["asym_id"]
        asym_id_atom = torch.bmm(
            feats["atom_to_token"].float(), asym_id_token.unsqueeze(-1).float()
        ).squeeze(-1)

    B, N, _ = pred_atom_coords.shape
    pred_atom_coords = pred_atom_coords.reshape(B // multiplicity, multiplicity, -1, 3)
    frames_idx_pred = (
        frames_idx_true.clone()
        .repeat_interleave(multiplicity, 0)
        .reshape(B // multiplicity, multiplicity, -1, 3)
    )

    # Iterate through the batch and modify the frames for nonpolymers
    for i, pred_atom_coord in enumerate(pred_atom_coords):
        token_idx = 0
        atom_idx = 0
        for id in torch.unique(asym_id_token[i]):
            mask_chain_token = (asym_id_token[i] == id) * feats["token_pad_mask"][i]
            mask_chain_atom = (asym_id_atom[i] == id) * feats["atom_pad_mask"][i]
            num_tokens = int(mask_chain_token.sum().item())
            num_atoms = int(mask_chain_atom.sum().item())
            if (
                feats["mol_type"][i, token_idx] != const.chain_type_ids["NONPOLYMER"]
                or num_atoms < 3
            ):
                token_idx += num_tokens
                atom_idx += num_atoms
                continue
            dist_mat = (
                (
                    pred_atom_coord[:, mask_chain_atom.bool()][:, None, :, :]
                    - pred_atom_coord[:, mask_chain_atom.bool()][:, :, None, :]
                )
                ** 2
            ).sum(-1) ** 0.5
            if inference:
                resolved_pair = 1 - (
                    feats["atom_pad_mask"][i][mask_chain_atom.bool()][None, :]
                    * feats["atom_pad_mask"][i][mask_chain_atom.bool()][:, None]
                ).to(torch.float32)
                resolved_pair[resolved_pair == 1] = torch.inf
                indices = torch.sort(dist_mat + resolved_pair, axis=2).indices
            else:
                if resolved_mask is None:
                    resolved_mask = feats["atom_resolved_mask"]
                resolved_pair = 1 - (
                    resolved_mask[i][mask_chain_atom.bool()][None, :]
                    * resolved_mask[i][mask_chain_atom.bool()][:, None]
                ).to(torch.float32)
                resolved_pair[resolved_pair == 1] = torch.inf
                indices = torch.sort(dist_mat + resolved_pair, axis=2).indices
            frames = (
                torch.cat(
                    [
                        indices[:, :, 1:2],
                        indices[:, :, 0:1],
                        indices[:, :, 2:3],
                    ],
                    dim=2,
                )
                + atom_idx
            )
            try:
                frames_idx_pred[i, :, token_idx : token_idx + num_atoms, :] = frames
            except Exception as e:
                print(f"Failed to process {feats['pdb_id']} due to {e}")
            token_idx += num_tokens
            atom_idx += num_atoms

    frames_expanded = pred_atom_coords[
        torch.arange(0, B // multiplicity, 1)[:, None, None, None].to(
            frames_idx_pred.device
        ),
        torch.arange(0, multiplicity, 1)[None, :, None, None].to(
            frames_idx_pred.device
        ),
        frames_idx_pred,
    ].reshape(-1, 3, 3)

    # Compute masks for collinearity / overlap
    mask_collinear_pred = compute_collinear_mask(
        frames_expanded[:, 1] - frames_expanded[:, 0],
        frames_expanded[:, 1] - frames_expanded[:, 2],
    ).reshape(B // multiplicity, multiplicity, -1)
    return frames_idx_pred, mask_collinear_pred * feats["token_pad_mask"][:, None, :]


def compute_aggregated_metric(logits, end=1.0):
    # Compute aggregated metric from logits
    num_bins = logits.shape[-1]
    bin_width = end / num_bins
    bounds = torch.arange(
        start=0.5 * bin_width, end=end, step=bin_width, device=logits.device
    )
    probs = nn.functional.softmax(logits, dim=-1)
    plddt = torch.sum(
        probs * bounds.view(*((1,) * len(probs.shape[:-1])), *bounds.shape),
        dim=-1,
    )
    return plddt


def tm_function(d, Nres):
    d0 = 1.24 * (torch.clip(Nres, min=19) - 15) ** (1 / 3) - 1.8
    return 1 / (1 + (d / d0) ** 2)


def compute_ptms(logits, x_preds, feats, multiplicity):
    # It needs to take as input the mask of the frames as they are not used to compute the PTM
    _, mask_collinear_pred = compute_frame_pred(
        x_preds, feats["frames_idx"], feats, multiplicity, inference=True
    )
    # mask overlapping, collinear tokens and ions (invalid frames)
    mask_pad = feats["token_pad_mask"].repeat_interleave(multiplicity, 0)
    maski = mask_collinear_pred.reshape(-1, mask_collinear_pred.shape[-1])
    pair_mask_ptm = maski[:, :, None] * mask_pad[:, None, :] * mask_pad[:, :, None]
    asym_id = feats["asym_id"].repeat_interleave(multiplicity, 0)
    pair_mask_iptm = (
        maski[:, :, None]
        * (asym_id[:, None, :] != asym_id[:, :, None])
        * mask_pad[:, None, :]
        * mask_pad[:, :, None]
    )
    num_bins = logits.shape[-1]
    bin_width = 32.0 / num_bins
    end = 32.0
    pae_value = torch.arange(
        start=0.5 * bin_width, end=end, step=bin_width, device=logits.device
    ).unsqueeze(0)
    N_res = mask_pad.sum(dim=-1, keepdim=True)
    tm_value = tm_function(pae_value, N_res).unsqueeze(1).unsqueeze(2)
    probs = nn.functional.softmax(logits, dim=-1)
    tm_expected_value = torch.sum(
        probs * tm_value,
        dim=-1,
    )  # shape (B, N, N)
    ptm = torch.max(
        torch.sum(tm_expected_value * pair_mask_ptm, dim=-1)
        / (torch.sum(pair_mask_ptm, dim=-1) + 1e-5),
        dim=1,
    ).values
    iptm = torch.max(
        torch.sum(tm_expected_value * pair_mask_iptm, dim=-1)
        / (torch.sum(pair_mask_iptm, dim=-1) + 1e-5),
        dim=1,
    ).values

    # compute ligand and protein iPTM
    token_type = feats["mol_type"]
    token_type = token_type.repeat_interleave(multiplicity, 0)
    is_ligand_token = (token_type == const.chain_type_ids["NONPOLYMER"]).float()
    is_protein_token = (token_type == const.chain_type_ids["PROTEIN"]).float()

    ligand_iptm_mask = (
        maski[:, :, None]
        * (asym_id[:, None, :] != asym_id[:, :, None])
        * mask_pad[:, None, :]
        * mask_pad[:, :, None]
        * (
            (is_ligand_token[:, :, None] * is_protein_token[:, None, :])
            + (is_protein_token[:, :, None] * is_ligand_token[:, None, :])
        )
    )
    protein_iptm_mask = (
        maski[:, :, None]
        * (asym_id[:, None, :] != asym_id[:, :, None])
        * mask_pad[:, None, :]
        * mask_pad[:, :, None]
        * (is_protein_token[:, :, None] * is_protein_token[:, None, :])
    )
    ligand_iptm = torch.max(
        torch.sum(tm_expected_value * ligand_iptm_mask, dim=-1)
        / (torch.sum(ligand_iptm_mask, dim=-1) + 1e-5),
        dim=1,
    ).values
    protein_iptm = torch.max(
        torch.sum(tm_expected_value * protein_iptm_mask, dim=-1)
        / (torch.sum(protein_iptm_mask, dim=-1) + 1e-5),
        dim=1,
    ).values

    # compute design and target iPTM and PTM
    if feats["design_mask"].sum()>0:
        is_chain_design_token = feats["chain_design_mask"].float()
        is_target_token = 1 - feats["chain_design_mask"].float()
    else:
        is_chain_design_token = feats["design_mask"].float()
        is_target_token = 1 - feats["design_mask"].float()
    design_iptm_mask = (
        maski[:, :, None]
        * mask_pad[:, None, :]
        * mask_pad[:, :, None]
        * (
            is_chain_design_token[:, :, None] * is_target_token[:, None, :]
            + is_target_token[:, :, None] * is_chain_design_token[:, None, :]
        )
    )
    design_iptm = torch.max(
        torch.sum(tm_expected_value * design_iptm_mask, dim=-1)
        / (torch.sum(design_iptm_mask, dim=-1) + 1e-5),
        dim=1,
    ).values

    is_design_token = feats["design_mask"].float()
    is_target_token = 1 - is_chain_design_token

    design_to_target_iptm_mask = (
        maski[:, :, None]
        * mask_pad[:, None, :]
        * mask_pad[:, :, None]
        * (
            is_design_token[:, :, None] * is_target_token[:, None, :]
            + is_target_token[:, :, None] * is_design_token[:, None, :]
        )
    )
    design_to_target_iptm = torch.max(
        torch.sum(tm_expected_value * design_to_target_iptm_mask, dim=-1)
        / (torch.sum(design_to_target_iptm_mask, dim=-1) + 1e-5),
        dim=1,
    ).values

    design_ptm_mask = (
        maski[:, :, None]
        * mask_pad[:, None, :]
        * mask_pad[:, :, None]
        * (is_chain_design_token[:, :, None] * is_chain_design_token[:, None, :])
    )
    target_ptm_mask = (
        maski[:, :, None]
        * mask_pad[:, None, :]
        * mask_pad[:, :, None]
        * (is_target_token[:, :, None] * is_target_token[:, None, :])
    )

    design_ptm = torch.max(
        torch.sum(tm_expected_value * design_ptm_mask, dim=-1)
        / (torch.sum(design_ptm_mask, dim=-1) + 1e-5),
        dim=1,
    ).values
    target_ptm = torch.max(
        torch.sum(tm_expected_value * target_ptm_mask, dim=-1)
        / (torch.sum(target_ptm_mask, dim=-1) + 1e-5),
        dim=1,
    ).values

    # Compute pair chain ipTM
    chain_pair_iptm = {}
    asym_ids_list = torch.unique(asym_id).tolist()
    for idx1 in asym_ids_list:
        chain_iptm = {}
        for idx2 in asym_ids_list:
            mask_pair_chain = (
                maski[:, :, None]
                * (asym_id[:, None, :] == idx1)
                * (asym_id[:, :, None] == idx2)
                * mask_pad[:, None, :]
                * mask_pad[:, :, None]
            )

            chain_iptm[idx2] = torch.max(
                torch.sum(tm_expected_value * mask_pair_chain, dim=-1)
                / (torch.sum(mask_pair_chain, dim=-1) + 1e-5),
                dim=1,
            ).values
        chain_pair_iptm[idx1] = chain_iptm

    # Compute mask of part of design that has an atom within X Angstrom of any target atom
    x_pred_half = x_preds.bfloat16()
    interact_mask = torch.zeros_like(feats["design_mask"])
    if (
        is_chain_design_token.sum() > 0
        and (~(is_chain_design_token.int())).sum() > 0
    ):
        atom_chain_design_mask = (
            torch.bmm(
                feats["atom_to_token"].bfloat16(),
                feats["chain_design_mask"].bfloat16().unsqueeze(-1),
            )
            .squeeze(-1)
            .bool()
        )
        x_target = x_pred_half[~atom_chain_design_mask]
        x_design = x_pred_half[atom_chain_design_mask]
        dists = torch.cdist(x_target, x_design)
        atom_interact_mask = torch.zeros_like(atom_chain_design_mask)
        atom_interact_mask[atom_chain_design_mask] = (dists.min(0)[0] < 8).unsqueeze(0)
        index = torch.argmax(feats["atom_to_token"].int(), dim=-1)
        interact_mask.scatter_(dim=1, index=index, src=atom_interact_mask)

    # Compute interface interaction pTM values: iipTM
    # Any ptm of the design that is interacting with the target to anything in the target
    is_interacting = interact_mask.float()
    is_target_token = 1 - is_chain_design_token.float()
    design_iiptm_mask = (
        maski[:, :, None]
        * mask_pad[:, None, :]
        * mask_pad[:, :, None]
        * (
            is_interacting[:, :, None] * is_target_token[:, None, :]
            + is_target_token[:, :, None] * is_interacting[:, None, :]
        )
    )
    design_iiptm = torch.max(
        torch.sum(tm_expected_value * design_iiptm_mask, dim=-1)
        / (torch.sum(design_iiptm_mask, dim=-1) + 1e-5),
        dim=1,
    ).values

    return (
        ptm,
        iptm,
        ligand_iptm,
        protein_iptm,
        chain_pair_iptm,
        design_to_target_iptm,
        design_iptm,
        design_iiptm,
        target_ptm,
        design_ptm,
    )


class GaussianSmearing(torch.nn.Module):
    def __init__(
        self,
        start: float = 0.0,
        stop: float = 5.0,
        num_gaussians: int = 50,
    ):
        super().__init__()
        offset = torch.linspace(start, stop, num_gaussians)
        self.num_gaussians = num_gaussians
        self.coeff = -0.5 / (offset[1] - offset[0]).item() ** 2
        self.register_buffer("offset", offset)

    def forward(self, dist):
        shape = dist.shape
        dist = dist.view(-1, 1) - self.offset.view(1, -1)
        return torch.exp(self.coeff * torch.pow(dist, 2)).reshape(
            *shape, self.num_gaussians
        )



# started from code from https://github.com/lucidrains/alphafold3-pytorch, MIT License, Copyright (c) 2024 Phil Wang
from functools import partial
from math import pi

import torch
from einops import rearrange
from torch import nn
from torch.nn import Linear, Module, ModuleList
from torch.nn.functional import one_hot

from boltzgen.data import const
import boltzgen.model.layers.initialize as init
from boltzgen.model.layers.transition import Transition
from boltzgen.model.modules.transformers import (
    AtomTransformer,
)
from boltzgen.model.modules.utils import (
    GaussianRandom3DEncodings,
    GaussianSmearing,
    LinearNoBias,
)


class FourierEmbedding(Module):
    """Algorithm 22."""

    def __init__(self, dim):
        super().__init__()
        self.proj = nn.Module()
        self.proj.register_buffer("weight", torch.randn(dim, 1))
        self.proj.register_buffer("bias", torch.randn(dim))

    def forward(
        self,
        times,  # Float[' b'],
    ):  # -> Float['b d']:
        times = rearrange(times, "b -> b 1")
        rand_proj = torch.addmm(self.proj.bias, times, self.proj.weight.t())
        return torch.cos(2 * pi * rand_proj)


class RelativePositionEncoder(Module):
    """Algorithm 3."""

    def __init__(self, token_z, r_max=32, s_max=2):
        super().__init__()
        self.r_max = r_max
        self.s_max = s_max
        self.linear_layer = LinearNoBias(4 * (r_max + 1) + 2 * (s_max + 1) + 1, token_z)

    def forward(self, feats):
        b_same_chain = torch.eq(
            feats["feature_asym_id"][:, :, None], feats["feature_asym_id"][:, None, :]
        )
        b_same_residue = torch.eq(
            feats["feature_residue_index"][:, :, None],
            feats["feature_residue_index"][:, None, :],
        )
        b_same_entity = torch.eq(
            feats["entity_id"][:, :, None], feats["entity_id"][:, None, :]
        )

        d_residue = (
            feats["feature_residue_index"][:, :, None]
            - feats["feature_residue_index"][:, None, :]
        )

        if torch.any(feats["cyclic"] > 0):
            period = torch.where(
                feats["cyclic"] > 0,
                feats["cyclic"],
                torch.zeros_like(feats["cyclic"]) + 10000,
            ).unsqueeze(1)
            d_residue = (d_residue - period * torch.round(d_residue / period)).long()

        d_residue = torch.clip(
            d_residue + self.r_max,
            0,
            2 * self.r_max,
        )
        d_residue = torch.where(
            b_same_chain,
            d_residue,
            torch.zeros_like(d_residue) + 2 * self.r_max + 1,
        )

        a_rel_pos = one_hot(d_residue, 2 * self.r_max + 2)

        d_token = torch.clip(
            feats["token_index"][:, :, None]
            - feats["token_index"][:, None, :]
            + self.r_max,
            0,
            2 * self.r_max,
        )
        d_token = torch.where(
            b_same_chain & b_same_residue,
            d_token,
            torch.zeros_like(d_token) + 2 * self.r_max + 1,
        )
        a_rel_token = one_hot(d_token, 2 * self.r_max + 2)

        d_chain = torch.clip(
            feats["sym_id"][:, :, None] - feats["sym_id"][:, None, :] + self.s_max,
            0,
            2 * self.s_max,
        )

        d_chain = torch.where(
            (~b_same_entity),
            torch.zeros_like(d_chain) + 2 * self.s_max + 1,
            d_chain,
        )
        # Note: added  | (~b_same_entity) based on observation of ProteinX manuscript
        a_rel_chain = one_hot(d_chain, 2 * self.s_max + 2)

        p = self.linear_layer(
            torch.cat(
                [
                    a_rel_pos.float(),
                    a_rel_token.float(),
                    b_same_entity.unsqueeze(-1).float(),
                    a_rel_chain.float(),
                ],
                dim=-1,
            )
        )
        return p


class SingleConditioning(Module):
    """Algorithm 21."""

    def __init__(
        self,
        sigma_data: float,
        tfmr_s: int = 768,
        token_s: int = 384,
        dim_fourier: int = 256,
        num_transitions: int = 2,
        transition_expansion_factor: int = 2,
        eps: float = 1e-20,
        disable_times: bool = False,
    ) -> None:
        super().__init__()
        self.eps = eps
        self.sigma_data = sigma_data
        self.disable_times = disable_times

        self.norm_single = nn.LayerNorm(token_s * 2)

        if tfmr_s != token_s * 2:
            self.token_s_to_tfmr_s = nn.Linear(token_s * 2, tfmr_s)
        else:
            self.token_s_to_tfmr_s = None

        self.single_embed = nn.Linear(tfmr_s, tfmr_s)
        if not self.disable_times:
            self.fourier_embed = FourierEmbedding(dim_fourier)
            self.norm_fourier = nn.LayerNorm(dim_fourier)
            self.fourier_to_single = LinearNoBias(dim_fourier, tfmr_s)

        transitions = ModuleList([])
        for _ in range(num_transitions):
            transition = Transition(
                dim=tfmr_s, hidden=transition_expansion_factor * tfmr_s
            )
            transitions.append(transition)

        self.transitions = transitions

    def forward(
        self,
        times,  # Float[' b'],
        s_trunk,  # Float['b n ts'],
        s_inputs,  # Float['b n ts'],
    ):  # -> Float['b n 2ts']:
        s = torch.cat((s_trunk, s_inputs), dim=-1)
        s = self.norm_single(s)

        if self.token_s_to_tfmr_s is not None:
            s = self.token_s_to_tfmr_s(s)

        s = self.single_embed(s)
        if not self.disable_times:
            fourier_embed = self.fourier_embed(
                times
            )  # note: sigma rescaling done in diffusion module
            normed_fourier = self.norm_fourier(fourier_embed)
            fourier_to_single = self.fourier_to_single(normed_fourier)

            s = rearrange(fourier_to_single, "b d -> b 1 d") + s

        for transition in self.transitions:
            s = transition(s) + s

        return s, normed_fourier if not self.disable_times else None


class PairwiseConditioning(Module):
    """Algorithm 21."""

    def __init__(
        self,
        token_z,
        dim_token_rel_pos_feats,
        num_transitions=2,
        transition_expansion_factor=2,
    ):
        super().__init__()

        self.dim_pairwise_init_proj = nn.Sequential(
            nn.LayerNorm(token_z + dim_token_rel_pos_feats),
            LinearNoBias(token_z + dim_token_rel_pos_feats, token_z),
        )

        transitions = ModuleList([])
        for _ in range(num_transitions):
            transition = Transition(
                dim=token_z, hidden=transition_expansion_factor * token_z
            )
            transitions.append(transition)

        self.transitions = transitions

    def forward(
        self,
        z_trunk,  # Float['b n n tz'],
        token_rel_pos_feats,  # Float['b n n 3'],
    ):  # -> Float['b n n tz']:
        z = torch.cat((z_trunk, token_rel_pos_feats), dim=-1)
        z = self.dim_pairwise_init_proj(z)

        for transition in self.transitions:
            z = transition(z) + z

        return z


class CoordinateConditioning(Module):
    def __init__(
        self,
        sigma_data: float,
        atom_s,
        token_s,
        num_heads,
        tfmr_s: int = 768,
        dim_fourier: int = 256,
        atom_feature_dim: int = 132,
        structure_prediction=True,
        disable_times: bool = False,
    ):
        super().__init__()
        self.sigma_data = sigma_data
        self.single_embed = LinearNoBias(token_s * 2, tfmr_s)

        self.disable_times = disable_times
        if not self.disable_times:
            self.fourier_embed = FourierEmbedding(dim_fourier)
            self.norm_fourier = nn.LayerNorm(dim_fourier)
            self.fourier_to_single = LinearNoBias(dim_fourier, tfmr_s)

        self.embed_atom_features = Linear(atom_feature_dim, num_heads)
        self.embed_atompair_ref_coord = LinearNoBias(3, num_heads)
        self.embed_atompair_ref_dist = LinearNoBias(1, num_heads)
        self.embed_atompair_mask = LinearNoBias(1, num_heads)

        self.structure_prediction = structure_prediction
        if structure_prediction:
            self.s_to_c_trans = nn.Sequential(
                nn.LayerNorm(tfmr_s), LinearNoBias(tfmr_s, 1)
            )
            init.final_init_(self.s_to_c_trans[1].weight)

    def forward(self, s_trunk, s_inputs, times, feats, atom_coords_noisy):
        s = torch.cat((s_trunk, s_inputs), dim=-1)
        s = self.single_embed(s)
        if not self.disable_times:
            fourier_embed = self.fourier_embed(
                times
            )  # note: sigma rescaling done in diffusion module
            normed_fourier = self.norm_fourier(fourier_embed)
            fourier_to_single = self.fourier_to_single(normed_fourier)
            s = rearrange(fourier_to_single, "b d -> b 1 d") + s

        with torch.autocast("cuda", enabled=False):
            atom_mask = feats["atom_pad_mask"].bool()
            atom_ref_pos = feats["ref_pos"]
            atom_uid = feats["ref_space_uid"]

            atom_feats = [
                atom_ref_pos,
                feats["ref_charge"].unsqueeze(-1),
                feats["ref_element"],
            ]

            atom_feats = torch.cat(atom_feats, dim=-1)
            c = self.embed_atom_features(atom_feats)
            B, N = atom_coords_noisy.shape[:2]

            s_to_c = self.s_to_c_trans(s.float())
            s_to_c = torch.bmm(
                feats["atom_to_token"].float().repeat_interleave(B, 0), s_to_c
            )
            c = c + s_to_c.to(c)

            atom_mask = atom_mask.repeat_interleave(B, 0)
            atom_uid = atom_uid.repeat_interleave(B, 0)

            d = atom_coords_noisy.view(B, N, 1, 3) - atom_coords_noisy.view(B, 1, N, 3)
            d_norm = torch.sum(d * d, dim=-1, keepdim=True)
            d_norm = 1 / (1 + d_norm)
            atom_mask_queries = atom_mask.view(B, N, 1)
            atom_mask_keys = atom_mask.view(B, 1, N)
            atom_uid_queries = atom_uid.view(B, N, 1)
            atom_uid_keys = atom_uid.view(B, 1, N)
            v = (
                (
                    atom_mask_queries
                    & atom_mask_keys
                    & (atom_uid_queries == atom_uid_keys)
                )
                .float()
                .unsqueeze(-1)
            )

            p = (
                self.embed_atompair_ref_coord(d) * v
                + self.embed_atompair_ref_dist(d_norm) * v
                + self.embed_atompair_mask(v) * v
            )
            p = p + c.view(B, 1, N, -1) + c.view(B, N, 1, -1)

            return p.sum(dim=0, keepdim=True)


class DistanceTokenEncoder(Module):
    def __init__(
        self,
        distance_gaussian_dim: int,
        token_z: int,
        out_dim: int,
    ):
        super().__init__()
        self.distance_gaussian_smearing = GaussianSmearing(
            start=0.0, stop=2.0, num_gaussians=distance_gaussian_dim
        )
        input_dim = distance_gaussian_dim + 1 + token_z
        self.distance_token_bias_trans = Transition(
            dim=input_dim, hidden=token_z, out_dim=out_dim
        )

    def forward(
        self,
        relative_position_encoding,
        feats,
    ):
        B, N, _, _ = relative_position_encoding.shape

        token_to_bb4_atoms = feats["token_to_bb4_atoms"]
        r = feats["coords"]

        r_repr = torch.bmm(
            token_to_bb4_atoms.float().view(B, N * 4, -1), r.view(B, -1, 3)
        )
        r_repr = r_repr.reshape(B, N, 4, 3).permute(0, 2, 1, 3)
        d = (r_repr.unsqueeze(-2) - r_repr.unsqueeze(-3)).norm(dim=-1).unsqueeze(-1)
        distance_gaussian = self.distance_gaussian_smearing(d)

        relative_position_encoding = relative_position_encoding.view(
            B, 1, N, N, -1
        ).expand(-1, 4, -1, -1, -1)
        distance_token_bias_input = torch.cat(
            (
                distance_gaussian,
                d,
                relative_position_encoding,
            ),
            dim=-1,
        )
        distance_token_bias = (
            self.distance_token_bias_trans(distance_token_bias_input)
            .permute(0, 2, 3, 4, 1)
            .reshape(B, N, N, -1)
        )
        return distance_token_bias


def get_indexing_matrix(K, W, H, device):
    assert W % 2 == 0
    assert H % (W // 2) == 0

    h = H // (W // 2)
    assert h % 2 == 0

    arange = torch.arange(2 * K, device=device)
    index = ((arange.unsqueeze(0) - arange.unsqueeze(1)) + h // 2).clamp(
        min=0, max=h + 1
    )
    index = index.view(K, 2, 2 * K)[:, 0, :]
    onehot = one_hot(index, num_classes=h + 2)[..., 1:-1].transpose(1, 0)
    return onehot.reshape(2 * K, h * K).float()


def single_to_keys(single, indexing_matrix, W, H):
    B, N, D = single.shape
    K = N // W
    single = single.view(B, 2 * K, W // 2, D)
    return torch.einsum("b j i d, j k -> b k i d", single, indexing_matrix).reshape(
        B, K, H, D
    )  # j = 2K, i = W//2, k = h * K


class AtomEncoder(Module):
    def __init__(
        self,
        atom_s,
        atom_z,
        token_s,
        token_z,
        atoms_per_window_queries,
        atoms_per_window_keys,
        atom_feature_dim,
        structure_prediction=True,
    ):
        super().__init__()

        self.embed_atom_features = Linear(atom_feature_dim, atom_s)
        self.embed_atompair_ref_pos = LinearNoBias(3, atom_z)
        self.embed_atompair_ref_dist = LinearNoBias(1, atom_z)
        self.embed_atompair_mask = LinearNoBias(1, atom_z)
        self.atoms_per_window_queries = atoms_per_window_queries
        self.atoms_per_window_keys = atoms_per_window_keys

        self.structure_prediction = structure_prediction
        if structure_prediction:
            self.s_to_c_trans = nn.Sequential(
                nn.LayerNorm(token_s), LinearNoBias(token_s, atom_s)
            )
            init.final_init_(self.s_to_c_trans[1].weight)

            self.z_to_p_trans = nn.Sequential(
                nn.LayerNorm(token_z), LinearNoBias(token_z, atom_z)
            )
            init.final_init_(self.z_to_p_trans[1].weight)

        self.c_to_p_trans_k = nn.Sequential(
            nn.ReLU(),
            LinearNoBias(atom_s, atom_z),
        )
        init.final_init_(self.c_to_p_trans_k[1].weight)

        self.c_to_p_trans_q = nn.Sequential(
            nn.ReLU(),
            LinearNoBias(atom_s, atom_z),
        )
        init.final_init_(self.c_to_p_trans_q[1].weight)

        self.p_mlp = nn.Sequential(
            nn.ReLU(),
            LinearNoBias(atom_z, atom_z),
            nn.ReLU(),
            LinearNoBias(atom_z, atom_z),
            nn.ReLU(),
            LinearNoBias(atom_z, atom_z),
        )
        init.final_init_(self.p_mlp[5].weight)

    def forward(
        self,
        feats,
        s_trunk=None,  # Float['bm n ts'],
        z=None,  # Float['bm n n tz'],
    ):
        with torch.autocast("cuda", enabled=False):
            B, N, _ = feats["ref_pos"].shape
            atom_mask = feats["atom_pad_mask"].bool()  # Bool['b m'],

            atom_ref_pos = feats["ref_pos"]  # Float['b m 3'],
            atom_uid = feats["ref_space_uid"]  # Long['b m'],

            atom_feats = [
                atom_ref_pos,
                feats["ref_charge"].unsqueeze(-1),
                feats["ref_element"],
                feats["ref_atom_name_chars"].reshape(B, N, 4 * 64),
            ]

            atom_feats = torch.cat(atom_feats, dim=-1)

            c = self.embed_atom_features(atom_feats)

        # note we are already creating the windows to make it more efficient
        W, H = self.atoms_per_window_queries, self.atoms_per_window_keys
        B, N = c.shape[:2]
        K = N // W
        keys_indexing_matrix = get_indexing_matrix(K, W, H, c.device)
        to_keys = partial(
            single_to_keys, indexing_matrix=keys_indexing_matrix, W=W, H=H
        )

        atom_ref_pos_queries = atom_ref_pos.view(B, K, W, 1, 3)
        atom_ref_pos_keys = to_keys(atom_ref_pos).view(B, K, 1, H, 3)

        d = atom_ref_pos_keys - atom_ref_pos_queries  # Float['b k w h 3']
        d_norm = torch.sum(d * d, dim=-1, keepdim=True)  # Float['b k w h 1']
        d_norm = 1 / (1 + d_norm)  # AF3 feeds in the reciprocal of the distance norm

        atom_mask_queries = atom_mask.view(B, K, W, 1)
        atom_mask_keys = (
            to_keys(atom_mask.unsqueeze(-1).float()).view(B, K, 1, H).bool()
        )
        atom_uid_queries = atom_uid.view(B, K, W, 1)
        atom_uid_keys = to_keys(atom_uid.unsqueeze(-1).float()).view(B, K, 1, H).long()
        v = (
            (atom_mask_queries & atom_mask_keys & (atom_uid_queries == atom_uid_keys))
            .float()
            .unsqueeze(-1)
        )  # Bool['b k w h 1']

        p = self.embed_atompair_ref_pos(d) * v
        p = p + self.embed_atompair_ref_dist(d_norm) * v
        p = p + self.embed_atompair_mask(v) * v

        q = c

        if self.structure_prediction:
            # run only in structure model not in initial encoding
            atom_to_token = feats["atom_to_token"].float()  # Long['b m n'],

            s_to_c = self.s_to_c_trans(s_trunk.float())
            s_to_c = torch.bmm(atom_to_token, s_to_c)
            c = c + s_to_c.to(c)

            atom_to_token_queries = atom_to_token.view(B, K, W, atom_to_token.shape[-1])
            atom_to_token_keys = to_keys(atom_to_token)
            z_to_p = self.z_to_p_trans(z.float())
            z_to_p = torch.einsum(
                "bijd,bwki,bwlj->bwkld",
                z_to_p,
                atom_to_token_queries,
                atom_to_token_keys,
            )
            p = p + z_to_p.to(p)

        p = p + self.c_to_p_trans_q(c.view(B, K, W, 1, c.shape[-1]))
        p = p + self.c_to_p_trans_k(to_keys(c).view(B, K, 1, H, c.shape[-1]))
        p = p + self.p_mlp(p)

        return q, c, p, to_keys


class AtomAttentionEncoder(Module):
    def __init__(
        self,
        atom_s,
        token_s,
        atoms_per_window_queries,
        atoms_per_window_keys,
        atom_encoder_depth=3,
        atom_encoder_heads=4,
        structure_prediction=True,
        activation_checkpointing=False,
        gaussian_random_3d_encoding_dim=0,
        transformer_post_layer_norm=False,
        tfmr_s=None,
        use_qk_norm=False,
    ):
        super().__init__()

        self.structure_prediction = structure_prediction
        if structure_prediction:
            self.gaussian_random_3d_encoding_dim = gaussian_random_3d_encoding_dim
            if gaussian_random_3d_encoding_dim > 0:
                self.encoding_3d = GaussianRandom3DEncodings(
                    gaussian_random_3d_encoding_dim
                )
            r_input_size = 3 + gaussian_random_3d_encoding_dim
            self.r_to_q_trans = LinearNoBias(r_input_size, atom_s)
            init.final_init_(self.r_to_q_trans.weight)

        self.atom_encoder = AtomTransformer(
            dim=atom_s,
            dim_single_cond=atom_s,
            attn_window_queries=atoms_per_window_queries,
            attn_window_keys=atoms_per_window_keys,
            depth=atom_encoder_depth,
            heads=atom_encoder_heads,
            activation_checkpointing=activation_checkpointing,
            post_layer_norm=transformer_post_layer_norm,
            use_qk_norm=use_qk_norm,
        )

        self.atom_to_token_trans = nn.Sequential(
            LinearNoBias(atom_s, tfmr_s if structure_prediction else token_s),
            nn.ReLU(),
        )
        self.atoms_per_window_queries = atoms_per_window_queries
        self.atoms_per_window_keys = atoms_per_window_keys

    def forward(
        self,
        feats,
        q,
        c,
        atom_enc_bias,
        to_keys,
        r=None,  # Float['bm m 3'],
        multiplicity=1,
    ):
        B, N, _ = feats["ref_pos"].shape
        atom_mask = feats["atom_pad_mask"].bool()  # Bool['b m'],

        if self.structure_prediction:
            # only here the multiplicity kicks in because we use the different positions r
            q = q.repeat_interleave(multiplicity, 0)

            r_input = r
            if self.gaussian_random_3d_encoding_dim > 0:
                r_input = torch.cat([r_input, self.encoding_3d(r)], dim=-1)

            r_to_q = self.r_to_q_trans(r_input)
            q = q + r_to_q

        c = c.repeat_interleave(multiplicity, 0)
        atom_mask = atom_mask.repeat_interleave(multiplicity, 0)

        q = self.atom_encoder(
            q=q,
            mask=atom_mask,
            c=c,
            bias=atom_enc_bias,
            multiplicity=multiplicity,
            to_keys=to_keys,
        )

        with torch.autocast("cuda", enabled=False):
            q_to_a = self.atom_to_token_trans(q).float()
            atom_to_token = feats["atom_to_token"].float()
            atom_to_token = atom_to_token.repeat_interleave(multiplicity, 0)
            atom_to_token_mean = atom_to_token / (
                atom_to_token.sum(dim=1, keepdim=True) + 1e-6
            )
            a = torch.bmm(atom_to_token_mean.transpose(1, 2), q_to_a)

        a = a.to(q)

        return a, q, c, to_keys


class AtomAttentionDecoder(Module):
    """Algorithm 6."""

    def __init__(
        self,
        atom_s,
        tfmr_s,
        attn_window_queries,
        attn_window_keys,
        atom_decoder_depth=3,
        atom_decoder_heads=4,
        activation_checkpointing=False,
        transformer_post_layer_norm=False,
        predict_res_type=False,
        use_qk_norm=False,
    ):
        super().__init__()
        self.predict_res_type = predict_res_type
        self.a_to_q_trans = LinearNoBias(tfmr_s, atom_s)
        init.final_init_(self.a_to_q_trans.weight)

        self.atom_decoder = AtomTransformer(
            dim=atom_s,
            dim_single_cond=atom_s,
            attn_window_queries=attn_window_queries,
            attn_window_keys=attn_window_keys,
            depth=atom_decoder_depth,
            heads=atom_decoder_heads,
            activation_checkpointing=activation_checkpointing,
            post_layer_norm=transformer_post_layer_norm,
            use_qk_norm=use_qk_norm,
        )

        self.atom_feat_to_atom_pos_update = nn.Sequential(
            nn.LayerNorm(atom_s), LinearNoBias(atom_s, 3)
        )
        init.final_init_(self.atom_feat_to_atom_pos_update[1].weight)

        if predict_res_type:
            self.res_type_predictor = Linear(atom_s, len(const.tokens))

    def forward(
        self,
        a,  # Float['bm n 2ts'],
        q,  # Float['bm m as'],
        c,  # Float['bm m as'],
        atom_dec_bias,  # Float['bm m m az'],
        feats,
        to_keys,
        multiplicity=1,
    ):
        with torch.autocast("cuda", enabled=False):
            atom_to_token = feats["atom_to_token"].float()
            atom_to_token = atom_to_token.repeat_interleave(multiplicity, 0)

            a_to_q = self.a_to_q_trans(a.float())
            a_to_q = torch.bmm(atom_to_token, a_to_q)

        q = q + a_to_q.to(q)
        atom_mask = feats["atom_pad_mask"]  # Bool['b m'],
        atom_mask = atom_mask.repeat_interleave(multiplicity, 0)

        q = self.atom_decoder(
            q=q,
            mask=atom_mask,
            c=c,
            bias=atom_dec_bias,
            multiplicity=multiplicity,
            to_keys=to_keys,
        )

        res_type = None
        s_feat = None
        if self.predict_res_type:
            idx = torch.argmax(feats["atom_to_token"].int(), dim=-1)
            mask = feats["atom_pad_mask"].repeat_interleave(multiplicity, 0)
            idx = idx.repeat_interleave(multiplicity, 0)
            src = q * mask[:, :, None]
            idx_expanded = idx.unsqueeze(-1).expand(-1, -1, q.size(-1))
            s_feat = torch.zeros(
                (q.shape[0], feats["res_type"].shape[1], q.shape[-1]),
                device=idx_expanded.device,
            )
            s_feat.scatter_add_(dim=1, index=idx_expanded, src=src)

        if self.predict_res_type and s_feat is not None:
            res_type = self.res_type_predictor(s_feat)


        r_update = self.atom_feat_to_atom_pos_update(q)
        return r_update, res_type



from torch import Tensor, nn

import boltzgen.model.layers.initialize as init


class Transition(nn.Module):
    """Perform a two-layer MLP."""

    def __init__(
        self,
        dim: int = 128,
        hidden: int = 512,
        out_dim: int = None,
    ) -> None:
        """Initialize the TransitionUpdate module.

        Parameters
        ----------
        dim: int
            The dimension of the input, default 128
        hidden: int
            The dimension of the hidden, default 512

        """
        super().__init__()
        if out_dim is None:
            out_dim = dim

        self.norm = nn.LayerNorm(dim, eps=1e-5)
        self.fc1 = nn.Linear(dim, hidden, bias=False)
        self.fc2 = nn.Linear(dim, hidden, bias=False)
        self.fc3 = nn.Linear(hidden, out_dim, bias=False)
        self.silu = nn.SiLU()
        self.hidden = hidden

        init.bias_init_one_(self.norm.weight)
        init.bias_init_zero_(self.norm.bias)

        init.lecun_normal_init_(self.fc1.weight)
        init.lecun_normal_init_(self.fc2.weight)
        init.final_init_(self.fc3.weight)

    def forward(self, x: Tensor, chunk_size: int = None) -> Tensor:
        """Perform a forward pass.

        Parameters
        ----------
        x: torch.Tensor
            The input data of shape (..., D)

        Returns
        -------
        x: torch.Tensor
            The output data of shape (..., D)

        """
        x = self.norm(x)

        if chunk_size is None or self.training:
            x = self.silu(self.fc1(x)) * self.fc2(x)
            x = self.fc3(x)
            return x
        else:
            for i in range(0, self.hidden, chunk_size):
                fc1_slice = self.fc1.weight[i : i + chunk_size, :]
                fc2_slice = self.fc2.weight[i : i + chunk_size, :]
                fc3_slice = self.fc3.weight[:, i : i + chunk_size]
                x_chunk = self.silu((x @ fc1_slice.T)) * (x @ fc2_slice.T)
                if i == 0:
                    x_out = x_chunk @ fc3_slice.T
                else:
                    x_out = x_out + x_chunk @ fc3_slice.T
            return x_out



from torch import Tensor
from typing import Dict
import torch

from boltzgen.data.mol import minimum_lddt_symmetry_coords
from boltzgen.model.loss.validation import weighted_minimum_rmsd


def get_true_coordinates(
    batch: Dict[str, Tensor],
    out: Dict[str, Tensor],
    diffusion_samples: int,
    symmetry_correction: bool,
    expand_to_diffusion_samples: bool = True,
    protein_lig_rmsd=False,
):
    if protein_lig_rmsd:
        assert not symmetry_correction, "Not implemented yet"

    if symmetry_correction:
        msg = "expand_to_diffusion_samples must be true for symmetry correction."
        assert expand_to_diffusion_samples, msg

    return_dict = {}

    if (
        symmetry_correction
    ):
        K = batch["coords"].shape[1]
        assert K == 1, (
            f"Symmetry correction is not supported for num_ensembles_val={K}."
        )

        assert batch["coords"].shape[0] == 1, (
            f"Validation is not supported for batch sizes={batch['coords'].shape[0]}"
        )

        true_coords = []
        true_coords_resolved_mask = []
        for idx in range(batch["token_index"].shape[0]):
            for rep in range(diffusion_samples):
                i = idx * diffusion_samples + rep

                best_true_coords, best_true_coords_resolved_mask = (
                    minimum_lddt_symmetry_coords(
                        coords=out["sample_atom_coords"][i : i + 1],
                        feats=batch,
                        index_batch=idx,
                    )
                )

                true_coords.append(best_true_coords)
                true_coords_resolved_mask.append(best_true_coords_resolved_mask)

        true_coords = torch.cat(true_coords, dim=0)
        true_coords_resolved_mask = torch.cat(true_coords_resolved_mask, dim=0)
        true_coords = true_coords.unsqueeze(1)


        return_dict["true_coords"] = true_coords
        return_dict["true_coords_resolved_mask"] = true_coords_resolved_mask
        return_dict["rmsds"] = 0
        return_dict["best_rmsd_recall"] = 0

    else:
        assert batch["coords"].shape[0] == 1, (
            f"Validation is not supported for batch sizes={batch['coords'].shape[0]}"
        )
        K, L = batch["coords"].shape[1:3]

        true_coords_resolved_mask = batch["atom_resolved_mask"] 
        true_coords = batch["coords"].squeeze(0)
        if expand_to_diffusion_samples:
            true_coords = true_coords.repeat((diffusion_samples, 1, 1)).reshape(
                diffusion_samples, K, L, 3
            )

            true_coords_resolved_mask = true_coords_resolved_mask.repeat_interleave(
                diffusion_samples, dim=0
            )  # since all masks are the same across conformers and diffusion samples, can just repeat S times
        else:
            true_coords_resolved_mask = true_coords_resolved_mask.squeeze(0)

        return_dict["true_coords"] = true_coords
        return_dict["true_coords_resolved_mask"] = true_coords_resolved_mask
        return_dict["rmsds"] = 0
        return_dict["best_rmsd_recall"] = 0
        return_dict["best_rmsd_precision"] = 0

        if protein_lig_rmsd:
            (
                rmsd,
                best_rmsd,
                rmsd_design,
                best_rmsd_design,
                rmsd_target,
                best_rmsd_target,
                rmsd_design_target,
                best_rmsd_design_target,
                target_aligned_rmsd_design,
                best_target_aligned_rmsd_design,
            ) = weighted_minimum_rmsd(
                out["sample_atom_coords"],
                batch,
                multiplicity=diffusion_samples,
                protein_lig_rmsd=protein_lig_rmsd,
            )
            return_dict["rmsd"] = rmsd
            return_dict["best_rmsd"] = best_rmsd
            return_dict["rmsd_design"] = rmsd_design
            return_dict["best_rmsd_design"] = best_rmsd_design
            return_dict["rmsd_target"] = rmsd_target
            return_dict["best_rmsd_target"] = best_rmsd_target
            return_dict["rmsd_design_target"] = rmsd_design_target
            return_dict["best_rmsd_design_target"] = best_rmsd_design_target
            return_dict["target_aligned_rmsd_design"] = target_aligned_rmsd_design
            return_dict["best_target_aligned_rmsd_design"] = best_target_aligned_rmsd_design
    return return_dict
